from functools import partial

from .. import query
from ..graph import Link, Nothing
from ..types import TypeRef, Sequence
from ..query import merge
from ..types import Any
from ..engine import Query
from ..expr.refs import RequirementsExtractor
from ..expr.core import to_expr, S, THIS
from ..expr.checker import check, fn_types
from ..expr.compiler import ExpressionCompiler


def _create_result_proc(engine_query, procs, options):
    def result_proc():
        sq_result = engine_query.result()
        return [[proc(this, sq_result, *opt_args)
                 for proc, opt_args in zip(procs, options)]
                for this in sq_result[THIS]]
    return result_proc


def _yield_options(query_field, graph_field):
    options = query_field.options or {}
    for option in graph_field.options:
        value = options.get(option.name, option.default)
        if value is Nothing:
            raise TypeError('Required option "{}" for {!r} was not provided'
                            .format(option.name, graph_field))
        else:
            yield value


class BoundExpr:

    def __init__(self, sub_graph, expr):
        self.sub_graph = sub_graph
        self.expr = expr

    def __repr__(self):
        expr, _ = to_expr(self.expr)
        return ('<{}: sub_graph={!r}, expr={!r}>'
                .format(self.__class__.__name__, self.sub_graph, expr))

    def __postprocess__(self, field):
        expr, funcs = to_expr(self.expr)

        env = fn_types(funcs)
        env.update(self.sub_graph.types['__root__'].__field_types__)
        env.update((opt.name, opt.type or Any) for opt in field.options)
        env[THIS] = TypeRef[self.sub_graph.node]

        expr = check(expr, self.sub_graph.types, env)

        option_names_set = set(opt.name for opt in field.options)
        reqs = RequirementsExtractor.extract(self.sub_graph.types, expr)
        reqs = query.Node([f for f in reqs.fields
                           if f.name not in option_names_set])

        option_names = [opt.name for opt in field.options]
        code = ExpressionCompiler.compile_lambda_expr(expr, option_names)
        proc = partial(eval(compile(code, '<expr>', 'eval')),
                       {f.__def_name__: f.__def_body__ for f in funcs})
        field.func = CheckedExpr(self.sub_graph, expr, reqs, proc)

    def __call__(self, *args, **kwargs):
        raise TypeError('Expression is not checked: {!r}'.format(self.expr))


class CheckedExpr:

    def __init__(self, sub_graph, expr, reqs, proc):
        self.sub_graph = sub_graph
        self.expr = expr
        self.reqs = reqs
        self.proc = proc

    def __repr__(self):
        return ('<{}: sub_graph={!r}, expr={!r}, reqs={!r}>'
                .format(self.__class__.__name__,
                        self.sub_graph, self.expr, self.reqs))

    @property
    def __subquery__(self):
        return self.sub_graph


class SubGraph:

    def __init__(self, graph, node):
        self.graph = graph
        self.node = node
        self.types = graph.__types__

    def __repr__(self):
        return '<{}: node={!r}>'.format(self.__class__.__name__, self.node)

    @property
    def __subquery__(self):
        return self

    def __postprocess__(self, field):
        BoundExpr(self, getattr(S.this, field.name)).__postprocess__(field)

    def __call__(self, fields, ids, queue, ctx, task_set):
        this_graph_link = Link(THIS, Sequence[TypeRef[self.node]], None,
                               requires=None)

        reqs = merge([gf.func.reqs for gf, _ in fields])
        procs = [gf.func.proc for gf, _ in fields]
        option_values = [[qf.options[opt.name] for opt in gf.options]
                         for gf, qf in fields]

        this_query_link = reqs.fields_map[THIS]
        other_reqs = query.Node([r for r in reqs.fields
                                 if r.name != THIS])

        q = Query(queue, task_set, self.graph, reqs, ctx)
        q.process_link(self.graph.root, this_graph_link, this_query_link,
                       None, ids)
        q.process_node(self.graph.root, other_reqs, None)
        return _create_result_proc(q, procs, option_values)

    def compile(self, expr):
        return BoundExpr(self, expr)

    def c(self, expr):
        return self.compile(expr)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hiku/protobuf/query.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='hiku/protobuf/query.proto',
  package='hiku.protobuf.query',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x19hiku/protobuf/query.proto\x12\x13hiku.protobuf.query\x1a\x1cgoogle/protobuf/struct.proto\"?\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12(\n\x07options\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"g\n\x04Link\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\'\n\x04node\x18\x02 \x01(\x0b\x32\x19.hiku.protobuf.query.Node\x12(\n\x07options\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"g\n\x04Item\x12+\n\x05\x66ield\x18\x01 \x01(\x0b\x32\x1a.hiku.protobuf.query.FieldH\x00\x12)\n\x04link\x18\x02 \x01(\x0b\x32\x19.hiku.protobuf.query.LinkH\x00\x42\x07\n\x05value\"0\n\x04Node\x12(\n\x05items\x18\x01 \x03(\x0b\x32\x19.hiku.protobuf.query.Itemb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_FIELD = _descriptor.Descriptor(
  name='Field',
  full_name='hiku.protobuf.query.Field',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hiku.protobuf.query.Field.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options', full_name='hiku.protobuf.query.Field.options', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=143,
)


_LINK = _descriptor.Descriptor(
  name='Link',
  full_name='hiku.protobuf.query.Link',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hiku.protobuf.query.Link.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node', full_name='hiku.protobuf.query.Link.node', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options', full_name='hiku.protobuf.query.Link.options', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=248,
)


_ITEM = _descriptor.Descriptor(
  name='Item',
  full_name='hiku.protobuf.query.Item',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='field', full_name='hiku.protobuf.query.Item.field', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link', full_name='hiku.protobuf.query.Item.link', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='hiku.protobuf.query.Item.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=250,
  serialized_end=353,
)


_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='hiku.protobuf.query.Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='hiku.protobuf.query.Node.items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=355,
  serialized_end=403,
)

_FIELD.fields_by_name['options'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LINK.fields_by_name['node'].message_type = _NODE
_LINK.fields_by_name['options'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_ITEM.fields_by_name['field'].message_type = _FIELD
_ITEM.fields_by_name['link'].message_type = _LINK
_ITEM.oneofs_by_name['value'].fields.append(
  _ITEM.fields_by_name['field'])
_ITEM.fields_by_name['field'].containing_oneof = _ITEM.oneofs_by_name['value']
_ITEM.oneofs_by_name['value'].fields.append(
  _ITEM.fields_by_name['link'])
_ITEM.fields_by_name['link'].containing_oneof = _ITEM.oneofs_by_name['value']
_NODE.fields_by_name['items'].message_type = _ITEM
DESCRIPTOR.message_types_by_name['Field'] = _FIELD
DESCRIPTOR.message_types_by_name['Link'] = _LINK
DESCRIPTOR.message_types_by_name['Item'] = _ITEM
DESCRIPTOR.message_types_by_name['Node'] = _NODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Field = _reflection.GeneratedProtocolMessageType('Field', (_message.Message,), dict(
  DESCRIPTOR = _FIELD,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Field)
  ))
_sym_db.RegisterMessage(Field)

Link = _reflection.GeneratedProtocolMessageType('Link', (_message.Message,), dict(
  DESCRIPTOR = _LINK,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Link)
  ))
_sym_db.RegisterMessage(Link)

Item = _reflection.GeneratedProtocolMessageType('Item', (_message.Message,), dict(
  DESCRIPTOR = _ITEM,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Item)
  ))
_sym_db.RegisterMessage(Item)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), dict(
  DESCRIPTOR = _NODE,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Node)
  ))
_sym_db.RegisterMessage(Node)


# @@protoc_insertion_point(module_scope)



from .query import Node, Field, Link


class Handle:

    def __init__(self, name=None, options=None, alias=None, items=None):
        self.__field_name__ = name
        self.__field_options__ = options
        self.__field_alias__ = alias
        self.__node_items__ = items

    def __getattr__(self, name):
        assert self.__field_name__ is None, self.__field_name__
        return self.__class__(name)

    def __getitem__(self, items):
        assert self.__node_items__ is None, self.__node_items__
        if isinstance(items, Handle):
            items = [items]
        else:
            assert isinstance(items, tuple)
        self.__node_items__ = items
        return self

    def __lshift__(self, other):
        assert isinstance(other, Handle), type(other)
        assert not self.__field_options__
        assert not self.__field_alias__
        assert not self.__node_items__
        return self.__class__(other.__field_name__,
                              other.__field_options__,
                              self.__field_name__,  # alias
                              other.__node_items__)

    def __call__(self, **options):
        assert self.__field_options__ is None, self.__field_options__
        self.__field_options__ = options
        return self


Q = Handle()


def build(items):
    """Builds a query

    Example:

    .. code-block:: python

        query = build([
            Q.foo,
            Q.bar(arg=42)[
                Q.baz,
            ],
            Q.aliased << Q.field,
        ])

    :param items: list of fields
    :return: ready to execute query
    """
    query_items = []
    for handle in items:
        assert handle.__field_name__ is not None
        if handle.__node_items__ is None:
            query_items.append(Field(handle.__field_name__,
                                     handle.__field_options__,
                                     handle.__field_alias__))
        else:
            query_items.append(Link(handle.__field_name__,
                                    build(handle.__node_items__),
                                    handle.__field_options__,
                                    handle.__field_alias__))
    return Node(query_items)



import sys


class cached_property:

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


def const(name):
    t = type(name, (object,), {})
    t.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')
    return t


def listify(func):
    def wrapper(*args, **kwargs):
        return list(func(*args, **kwargs))
    return wrapper



"""
    hiku.readers.protobuf
    ~~~~~~~~~~~~~~~~~~~~~

    Support for queries encoded using Protocol Buffers

"""
from google.protobuf.json_format import MessageToDict

from ..query import Node, Link, Field, merge
from ..protobuf import query_pb2


def _transform(pb_node):
    fields = []
    for i in pb_node.items:
        item_type = i.WhichOneof('value')
        if item_type == 'field':
            if not i.field.name:
                raise TypeError('Field name is empty: {!r}'.format(i))
            options = None
            if i.field.HasField('options'):
                options = MessageToDict(i.field.options)
            fields.append(Field(i.field.name, options))
        elif item_type == 'link':
            if not i.link.name:
                raise TypeError('Link name is empty: {!r}'.format(i))
            options = None
            if i.link.HasField('options'):
                options = MessageToDict(i.link.options)
            fields.append(Link(i.link.name, _transform(i.link.node), options))
        else:
            raise TypeError('Node item is empty: {!r}'.format(i))
    return Node(fields)


def transform(pb_node):
    return merge([_transform(pb_node)])


def read(data):
    """Reads a query, encoded into binary Protocol Buffers format, using
    message types from the ``hiku.protobuf.query`` package.

    Proto-file location: ``hiku/protobuf/query.proto``

    Generated message types: ``hiku.protobuf.query_pb2``

    Example:

    .. code-block:: python

        from hiku.builder import Q, build
        from hiku.export.protobuf import export

        bin_query = export(build([
            Q.characters(limit=100)[
                Q.name,
            ],
        ])).SerializeToString()

        assert bin_query == (
            b'\\n(\\x12&\\n\\ncharacters\\x12\\n\\n\\x08\\n\\x06\\n\\x04'
            b'name\\x1a\\x0c\\n\\x05limit\\x12\\x03\\x10\\xc8\\x01'
        )

        query = read(bin_query)  # reading binary message

        result = engine.execute(graph, query)

    :param bytes data: binary message representation
    :return: :py:class:`hiku.query.Node`, ready to execute query object
    """
    pb_value = query_pb2.Node.FromString(data)
    return transform(pb_value)



from ..types import TypingMeta


class TypeDefMeta(TypingMeta):

    def __cls_init__(cls, params):
        cls.__type_name__, cls.__type__ = params

    def accept(cls, visitor):
        return visitor.visit_typedef(cls)


class TypeDef(metaclass=TypeDefMeta):
    pass



from abc import abstractmethod
from asyncio import gather
from contextlib import contextmanager
from typing import (
    List,
    Dict,
    Any,
    Optional,
    Tuple,
    overload,
    Iterator,
)

from .utils import get_keys
from .introspection import (
    FederatedGraphQLIntrospection,
    AsyncFederatedGraphQLIntrospection,
    is_introspection_query,
    extend_with_federation,
)
from .validate import validate

from hiku.denormalize.graphql import DenormalizeGraphQL
from hiku.federation.denormalize import DenormalizeEntityGraphQL
from hiku.endpoint.graphql import (
    BaseGraphQLEndpoint,
    _type_names,
    _switch_graph,
    GraphQLError,
    _StripQuery,
)
from hiku.graph import Graph
from hiku.query import Node
from hiku.result import Proxy, Reference
from hiku.readers.graphql import Operation


def _process_query(graph: Graph, query: Node) -> Node:
    stripped_query = _StripQuery().visit(query)
    errors = validate(graph, stripped_query)
    if errors:
        raise GraphQLError(errors=errors)
    else:
        return stripped_query


def denormalize_entities(
    graph: Graph,
    query: Node,
    result: Proxy,
) -> List[Dict[str, Any]]:

    entities_link = query.fields_map['_entities']
    node = entities_link.node
    representations = entities_link.options['representations']

    entities = []
    for r in representations:
        typename = r['__typename']
        for key in get_keys(graph, typename):
            if key not in r:
                continue
            ident = r[key]
            result.__ref__ = Reference(typename, ident)
            data = DenormalizeEntityGraphQL(
                graph, result, typename
            ).process(node)
            entities.append(data)

    return entities


class BaseFederatedGraphEndpoint(BaseGraphQLEndpoint):
    @abstractmethod
    def execute(
        self, graph: Graph, op: Operation, ctx: Optional[Dict]
    ) -> Dict:
        pass

    @abstractmethod
    def dispatch(self, data: Dict) -> Dict:
        pass

    @contextmanager
    def context(self, op: Operation) -> Iterator[Dict]:
        yield {}

    @staticmethod
    def postprocess_result(result: Proxy, graph: Graph, op: Operation) -> Dict:
        if '_service' in op.query.fields_map:
            return {'_service': {'sdl': result['sdl']}}
        elif '_entities' in op.query.fields_map:
            return {
                '_entities': denormalize_entities(
                    graph, op.query, result
                )
            }

        type_name = _type_names[op.type]

        data = DenormalizeGraphQL(graph, result, type_name).process(op.query)
        if is_introspection_query(op.query):
            extend_with_federation(graph, data)
        return data


class FederatedGraphQLEndpoint(BaseFederatedGraphEndpoint):
    """Can execute either regular or federated queries.
    Handles following fields of federated query:
        - _service
        - _entities
    """
    introspection_cls = FederatedGraphQLIntrospection

    def execute(
        self, graph: Graph, op: Operation, ctx: Optional[Dict]
    ) -> Dict:
        stripped_query = _process_query(graph, op.query)
        result = self.engine.execute(graph, stripped_query, ctx)
        return self.postprocess_result(result, graph, op)

    def dispatch(self, data: Dict) -> Dict:
        try:
            graph, op = _switch_graph(
                data, self.query_graph, self.mutation_graph,
            )
            with self.context(op) as ctx:
                result = self.execute(graph, op, ctx)
            return {'data': result}
        except GraphQLError as e:
            return {'errors': [{'message': e} for e in e.errors]}


class AsyncFederatedGraphQLEndpoint(BaseFederatedGraphEndpoint):
    introspection_cls = AsyncFederatedGraphQLIntrospection

    async def execute(  # type: ignore[override]
        self, graph: Graph, op: Operation, ctx: Optional[Dict]
    ) -> Dict:
        stripped_query = _process_query(graph, op.query)
        result = await self.engine.execute_async(graph, stripped_query, ctx)
        return self.postprocess_result(result, graph, op)

    async def dispatch(self, data: Dict) -> Dict:  # type: ignore[override]
        try:
            graph, op = _switch_graph(
                data, self.query_graph, self.mutation_graph,
            )

            with self.context(op) as ctx:
                result = await self.execute(graph, op, ctx)
            return {'data': result}
        except GraphQLError as e:
            return {'errors': [{'message': e} for e in e.errors]}


class AsyncBatchFederatedGraphQLEndpoint(AsyncFederatedGraphQLEndpoint):
    @overload  # type: ignore[override]
    async def dispatch(self, data: List[Dict]) -> Tuple[Dict, ...]:
        ...

    @overload
    async def dispatch(self, data: Dict) -> Dict:
        ...

    async def dispatch(self, data):  # type: ignore[no-untyped-def]
        if isinstance(data, list):
            return await gather(*(
                super().dispatch(item)
                for item in data
            ))

        return await super().dispatch(data)



"""
    hiku.result
    ~~~~~~~~~~~

    In all examples query results are showed in **denormalized** form, suitable
    for reading (by humans) and for serializing into simple formats, into `JSON`
    for example. But this is not how `Hiku` stores result internally.

    Internally `Hiku` stores result in a fully **normalized** form. So result in
    `Hiku` is also a graph structure with references between objects. This
    approach has lots of advantages:

      - normalization helps to heavily reduce size of serialized result when we
        need to transfer it (this avoids data duplication)
      - it reduces internal memory usage and simplifies work with data
        internally
      - gives ability to cache, precisely and effortlessly update local state
        on the client

"""
from collections import defaultdict

from .types import RecordMeta, OptionalMeta, SequenceMeta, get_type
from .query import Node, Field, Link
from .graph import Link as GraphLink, Field as GraphField, Many, Maybe
from .utils import cached_property


class Reference:
    __slots__ = ('node', 'ident')

    def __init__(self, node, ident):
        self.node = node
        self.ident = ident

    def __repr__(self):
        return '<{}[{!r}]>'.format(self.node, self.ident)


ROOT = Reference('__root__', '__root__')


class Index(defaultdict):

    def __init__(self):
        super(Index, self).__init__(lambda: defaultdict(dict))

    @cached_property
    def root(self):
        return self[ROOT.node][ROOT.ident]

    def finish(self):
        for value in self.values():
            value.default_factory = None
        self.default_factory = None


class Proxy:
    __slots__ = ('__idx__', '__ref__', '__node__')

    def __init__(self, index, reference, node):
        self.__idx__ = index
        self.__ref__ = reference
        self.__node__ = node

    def __getitem__(self, item):
        try:
            field = self.__node__.result_map[item]
        except KeyError:
            raise KeyError("Field {!r} wasn't requested in the query"
                           .format(item))

        try:
            obj = self.__idx__[self.__ref__.node][self.__ref__.ident]
        except KeyError:
            raise AssertionError('Object {}[{!r}] is missing in the index'
                                 .format(self.__ref__.node, self.__ref__.ident))
        try:
            value = obj[field.index_key]
        except KeyError:
            raise AssertionError('Field {}[{!r}].{} is missing in the index'
                                 .format(self.__ref__.node, self.__ref__.ident,
                                         field.index_key))

        if isinstance(field, Field):
            return value
        elif isinstance(value, Reference):
            return self.__class__(self.__idx__, value, field.node)
        elif (
            isinstance(value, list) and value
            and isinstance(value[0], Reference)
        ):
            return [self.__class__(self.__idx__, ref, field.node)
                    for ref in value]
        else:
            return value

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(e)


def _denormalize_type(type_, result, query_obj):
    if isinstance(query_obj, Field):
        return result
    elif isinstance(query_obj, Link):
        if isinstance(type_, SequenceMeta):
            return [_denormalize_type(type_.__item_type__, item, query_obj)
                    for item in result]
        elif isinstance(type_, OptionalMeta):
            return (_denormalize_type(type_.__type__, result, query_obj)
                    if result is not None else None)
        else:
            assert isinstance(type_, RecordMeta), type(type_)
            field_types = type_.__field_types__
            return {
                f.name: _denormalize_type(field_types[f.name],
                                          result[f.name], f)
                for f in query_obj.node.fields
            }
    assert False, (type_, query_obj)


def _denormalize(graph, graph_obj, result, query_obj):
    if isinstance(query_obj, Node):
        return {
            f.result_key: _denormalize(graph, graph_obj.fields_map[f.name],
                                       result[f.result_key], f)
            for f in query_obj.fields
        }

    elif isinstance(query_obj, Field):
        return result

    elif isinstance(query_obj, Link):
        if isinstance(graph_obj, GraphField):
            field_type = get_type(graph.data_types, graph_obj.type)
            return _denormalize_type(field_type, result, query_obj)

        elif isinstance(graph_obj, GraphLink):
            graph_node = graph.nodes_map[graph_obj.node]
            if graph_obj.type_enum is Many:
                return [_denormalize(graph, graph_node, v, query_obj.node)
                        for v in result]
            elif graph_obj.type_enum is Maybe and result is None:
                return None
            else:
                return _denormalize(graph, graph_node, result, query_obj.node)

        else:
            return _denormalize(graph, graph_obj, result, query_obj.node)


def denormalize(graph, result):
    """Transforms normalized result (graph) into simple hierarchical structure

    This hierarchical structure will follow query structure.

    .. note:: To work properly this function requires that incoming query
        was merged by using :py:func:`hiku.query.merge` function. This is done
        by default in all query readers.

    Example:

    .. code-block:: python

        query = hiku.readers.simple.read('[:foo]')
        norm_result = hiku_engine.execute(graph, query)
        result = hiku.result.denormalize(graph, norm_result)
        assert result == {'foo': 'value'}

    :param graph: :py:class:`~hiku.graph.Graph` definition
    :param result: result of the query
    """
    return _denormalize(graph, graph.root, result, result.__node__)



import sys
import ast as _ast


PY38 = sys.version_info >= (3, 8)


class _AST:

    def __getattr__(self, name):
        return getattr(_ast, name)

    if PY38:
        arguments = _ast.arguments
    else:
        @staticmethod
        def arguments(_, args, vararg, kwonlyargs, kw_defaults, kwarg, defaults):  # noqa
            return _ast.arguments(args, vararg, kwonlyargs, kw_defaults, kwarg, defaults)  # noqa


ast = _AST()



from abc import ABC, abstractmethod
from asyncio import gather

from ..graph import apply
from ..query import QueryTransformer
from ..validate.query import validate
from ..readers.graphql import read_operation, OperationType
from ..denormalize.graphql import DenormalizeGraphQL
from ..introspection.graphql import AsyncGraphQLIntrospection, QUERY_ROOT_NAME
from ..introspection.graphql import GraphQLIntrospection, MUTATION_ROOT_NAME


_type_names = {
    OperationType.QUERY: QUERY_ROOT_NAME,
    OperationType.MUTATION: MUTATION_ROOT_NAME,
}


class GraphQLError(Exception):

    def __init__(self, *, errors):
        super().__init__('{} errors'.format(len(errors)))
        self.errors = errors


class _StripQuery(QueryTransformer):

    def visit_node(self, obj):
        return obj.copy(fields=[self.visit(f) for f in obj.fields
                                if f.name != '__typename'])


def _switch_graph(data, query_graph, mutation_graph=None):
    try:
        op = read_operation(data['query'],
                            variables=data.get('variables'),
                            operation_name=data.get('operationName'))
    except TypeError as e:
        raise GraphQLError(errors=[
            'Failed to read query: {}'.format(e),
        ])
    if op.type is OperationType.QUERY:
        graph = query_graph
    elif op.type is OperationType.MUTATION and mutation_graph is not None:
        graph = mutation_graph
    else:
        raise GraphQLError(errors=[
            'Unsupported operation type: {!r}'.format(op.type),
        ])
    return graph, op


def _process_query(graph, query):
    stripped_query = _StripQuery().visit(query)
    errors = validate(graph, stripped_query)
    if errors:
        raise GraphQLError(errors=errors)
    else:
        return stripped_query


class BaseGraphQLEndpoint(ABC):

    @property
    @abstractmethod
    def introspection_cls(self):
        pass

    def __init__(self, engine, query_graph, mutation_graph=None):
        self.engine = engine

        introspection = self.introspection_cls(query_graph, mutation_graph)
        self.query_graph = apply(query_graph, [introspection])
        if mutation_graph is not None:
            self.mutation_graph = apply(mutation_graph, [introspection])
        else:
            self.mutation_graph = None

    @abstractmethod
    def execute(self, graph, op, ctx):
        pass

    @abstractmethod
    def dispatch(self, data):
        pass


class GraphQLEndpoint(BaseGraphQLEndpoint):
    introspection_cls = GraphQLIntrospection

    def execute(self, graph, op, ctx):
        stripped_query = _process_query(graph, op.query)
        result = self.engine.execute(graph, stripped_query, ctx)
        type_name = _type_names[op.type]
        return DenormalizeGraphQL(graph, result, type_name).process(op.query)

    def dispatch(self, data):
        try:
            graph, op = _switch_graph(
                data, self.query_graph, self.mutation_graph,
            )
            result = self.execute(graph, op, {})
            return {'data': result}
        except GraphQLError as e:
            return {'errors': [{'message': e} for e in e.errors]}


class BatchGraphQLEndpoint(GraphQLEndpoint):

    def dispatch(self, data):
        if isinstance(data, list):
            return [
                super(BatchGraphQLEndpoint, self).dispatch(item)
                for item in data
            ]
        else:
            return super(BatchGraphQLEndpoint, self).dispatch(data)


class AsyncGraphQLEndpoint(BaseGraphQLEndpoint):
    introspection_cls = AsyncGraphQLIntrospection

    async def execute(self, graph, op, ctx):
        stripped_query = _process_query(graph, op.query)
        result = await self.engine.execute(graph, stripped_query, ctx)
        type_name = _type_names[op.type]
        return DenormalizeGraphQL(graph, result, type_name).process(op.query)

    async def dispatch(self, data):
        try:
            graph, op = _switch_graph(
                data, self.query_graph, self.mutation_graph,
            )
            result = await self.execute(graph, op, {})
            return {'data': result}
        except GraphQLError as e:
            return {'errors': [{'message': e} for e in e.errors]}


class AsyncBatchGraphQLEndpoint(AsyncGraphQLEndpoint):

    async def dispatch(self, data):
        if isinstance(data, list):
            return await gather(*(
                super(AsyncBatchGraphQLEndpoint, self).dispatch(item)
                for item in data
            ))
        else:
            return await super(AsyncBatchGraphQLEndpoint, self).dispatch(data)



import typing as t

from math import isfinite
from typing import (
    Optional,
    Iterable,
    List,
)

from graphql.language.printer import print_ast
from graphql.language import ast
from graphql.pyutils import inspect

from hiku.graph import (
    Link,
    Nothing,
    GraphVisitor,
    Field,
    Node,
    Root,
    GraphTransformer,
    Graph,
    Option,
)
from hiku.types import (
    IntegerMeta,
    TypeRefMeta,
    StringMeta,
    SequenceMeta,
    OptionalMeta,
    AnyMeta,
    FloatMeta,
    BooleanMeta,
    Record,
)


def _name(value):
    return ast.NameNode(value=value) if value is not None else None


def coerce_type(x):
    if isinstance(x, str):
        return _name(x)
    return x


def _non_null_type(val):
    return ast.NonNullTypeNode(type=coerce_type(val))


def _encode_type(value) -> ast.TypeNode:
    def _encode(val):
        if isinstance(val, OptionalMeta):
            return _encode(val.__type__), True
        elif isinstance(val, TypeRefMeta):
            return val.__type_name__
        elif isinstance(val, IntegerMeta):
            return 'Int'
        elif isinstance(val, StringMeta):
            return 'String'
        elif isinstance(val, BooleanMeta):
            return 'Boolean'
        elif isinstance(val, SequenceMeta):
            return ast.ListTypeNode(type=_encode_type(val.__item_type__))
        elif isinstance(val, AnyMeta):
            return 'Any'
        elif isinstance(val, FloatMeta):
            return 'Float'
        elif val is None:
            return ''
        else:
            raise TypeError('Unsupported type: {!r}'.format(val))

    encoded = _encode(value)
    if isinstance(encoded, tuple):
        [type_, optional] = encoded
        if optional:
            return coerce_type(type_)
        return _non_null_type(type_)

    return _non_null_type(encoded)


def _encode_default_value(value) -> Optional[ast.ValueNode]:
    if value == Nothing:
        return None

    if value is None:
        return ast.NullValueNode()

    if isinstance(value, bool):
        return ast.BooleanValueNode(value=value)

    if isinstance(value, int):
        return ast.IntValueNode(value=f"{value:d}")
    if isinstance(value, float) and isfinite(value):
        return ast.FloatValueNode(value=f"{value:g}")

    if isinstance(value, str):
        return ast.StringValueNode(value=value)

    if isinstance(value, Iterable) and not isinstance(value, str):
        maybe_value_nodes = (_encode_default_value(item) for item in value)
        value_nodes: t.List[ast.ValueNode] = list(
            filter(None, maybe_value_nodes)
        )
        return ast.ListValueNode(values=[value_nodes])

    raise TypeError(f"Cannot convert value to AST: {inspect(value)}.")


class Exporter(GraphVisitor):
    def export_record(self, type_name: str, obj: Record):
        def new_field(name: str, type_):
            return ast.FieldDefinitionNode(
                name=_name(name),
                type=_encode_type(type_),
            )
        fields = [new_field(f_name, field)
                  for f_name, field in obj.__field_types__.items()]
        return ast.ObjectTypeDefinitionNode(
            name=_name(type_name),
            fields=fields,
        )

    def visit_graph(self, obj: Graph) -> List[ast.DefinitionNode]:
        """List of ObjectTypeDefinitionNode and ObjectTypeExtensionNode"""
        return [
            self.get_any_type(),
            *[self.export_record(type_name, type_)
              for type_name, type_ in obj.data_types.items()],
            *[self.visit(item) for item in obj.items]
        ]

    def get_any_type(self):
        return ast.ScalarTypeDefinitionNode(name=_name('Any'))

    def visit_root(self, obj: Root):
        return ast.ObjectTypeExtensionNode(
            name=_name('Query'),
            fields=[self.visit(item) for item in obj.fields],
        )

    def visit_field(self, obj: Field):
        return ast.FieldDefinitionNode(
            name=_name(obj.name),
            type=_encode_type(obj.type),
            arguments=[self.visit(o) for o in obj.options],
            directives=[self.visit(d) for d in obj.directives]
        )

    def visit_node(self, obj: Node):
        fields = [self.visit(field) for field in obj.fields]

        return ast.ObjectTypeDefinitionNode(
            name=_name(obj.name),
            fields=fields,
            directives=[self.visit(d) for d in obj.directives]
        )

    def visit_link(self, obj: Link):
        return ast.FieldDefinitionNode(
            name=_name(obj.name),
            arguments=[self.visit(o) for o in obj.options],
            type=_encode_type(obj.type),
            directives=[self.visit(d) for d in obj.directives]
        )

    def visit_option(self, obj: Option):
        return ast.InputValueDefinitionNode(
            name=_name(obj.name),
            description=obj.description,
            type=_encode_type(obj.type),
            default_value=_encode_default_value(obj.default),
        )

    def visit_key_directive(self, obj):
        return ast.DirectiveNode(
            name=_name('key'),
            arguments=[
                ast.ArgumentNode(
                    name=_name('fields'),
                    value=ast.StringValueNode(value=obj.fields),
                ),
            ],
        )

    def visit_provides_directive(self, obj):
        return ast.DirectiveNode(
            name=_name('provides'),
            arguments=[
                ast.ArgumentNode(
                    name=_name('fields'),
                    value=ast.StringValueNode(value=obj.fields),
                ),
            ],
        )

    def visit_requires_directive(self, obj):
        return ast.DirectiveNode(
            name=_name('requires'),
            arguments=[
                ast.ArgumentNode(
                    name=_name('fields'),
                    value=ast.StringValueNode(value=obj.fields),
                ),
            ],
        )

    def visit_external_directive(self, obj):
        return ast.DirectiveNode(name=_name('external'))

    def visit_extends_directive(self, obj):
        return ast.DirectiveNode(name=_name('extends'))

    def visit_deprecated_directive(self, obj):
        return ast.DirectiveNode(
            name=_name('deprecated'),
            arguments=[
                ast.ArgumentNode(
                    name=_name('reason'),
                    value=ast.StringValueNode(value=obj.reason),
                ),
            ],
        )


def get_ast(graph: Graph) -> ast.DocumentNode:
    graph = _StripGraph().visit(graph)
    return ast.DocumentNode(definitions=Exporter().visit(graph))


class _StripGraph(GraphTransformer):
    def visit_root(self, obj):
        def skip(field):
            return field.name in ['__typename', '_entities']

        return Root([self.visit(f) for f in obj.fields if not skip(f)])

    def visit_graph(self, obj):
        def skip(node):
            if node.name is None:
                # check if it is a Root node from introspection
                return '__schema' in node.fields_map

            return node.name.startswith('__')

        return Graph(
            [self.visit(node) for node in obj.items if not skip(node)],
            obj.data_types
        )

    def visit_node(self, obj):
        def skip(field):
            return field.name in ['__typename']

        return Node(
            obj.name,
            fields=[self.visit(f) for f in obj.fields if not skip(f)],
            description=obj.description,
            directives=obj.directives
        )


def print_sdl(graph: Graph) -> str:
    """Print graphql AST into a string"""
    return print_ast(get_ast(graph))



from itertools import chain
from contextlib import contextmanager
from collections import Counter

from ..directives import Deprecated
from ..graph import (
    GraphVisitor,
    Root,
    Field,
)
from ..graph import AbstractNode, AbstractField, AbstractLink, AbstractOption

from .errors import Errors


class GraphValidationError(TypeError):

    def __init__(self, errors):
        self.errors = errors
        errors_list = '\n'.join('- {}'.format(e) for e in errors)
        super(GraphValidationError, self).__init__('\n' + errors_list)


class GraphValidator(GraphVisitor):

    class _NameFormatter(GraphVisitor):

        def visit_node(self, obj):
            return obj.name

        def visit_root(self, obj):
            return 'root'

        def visit_link(self, obj):
            return '.{}'.format(obj.name)

        def visit_field(self, obj):
            return '.{}'.format(obj.name)

        def visit_option(self, obj):
            return ':{}'.format(obj.name)

    _name_formatter = _NameFormatter()
    _graph_accept_types = (AbstractNode,)
    _node_accept_types = (AbstractField, AbstractLink)
    _link_accept_types = (AbstractOption,)

    def __init__(self, items):
        self.items = items
        self.errors = Errors()
        self._ctx = []

    @classmethod
    def validate(cls, items):
        validator = cls(items)
        validator.visit_graph_items(items)
        if validator.errors.list:
            raise GraphValidationError(validator.errors.list)

    @contextmanager
    def push_ctx(self, obj):
        self._ctx.append(obj)
        try:
            yield
        finally:
            self._ctx.pop()

    @property
    def ctx(self):
        return self._ctx[-1]

    def _get_duplicates(self, names):
        counter = Counter(names)
        return [k for k, v in counter.items() if v > 1]

    def _format_names(self, names):
        return ', '.join('"{}"'.format(name) for name in names)

    def _format_types(self, objects):
        return ', '.join(map(repr, set(type(obj) for obj in objects)))

    def _format_path(self, obj=None):
        path = self._ctx + ([obj] if obj is not None else [])
        return ''.join(self._name_formatter.visit(i) for i in path)

    def _validate_deprecated_duplicates(self, obj):
        deprecated_count = sum(
            (1 for d in obj.directives if isinstance(d, Deprecated)))
        if deprecated_count > 1:
            self.errors.report(
                'Deprecated directive must be used only once for "{}", found {}'
                .format(self._format_path(obj), deprecated_count)
            )

    def visit_option(self, obj):
        # TODO: check option default value according to the option type
        pass

    def visit_field(self, obj: Field):
        self._validate_deprecated_duplicates(obj)

    def visit_link(self, obj):
        invalid = [f for f in obj.options
                   if not isinstance(f, self._link_accept_types)]
        if invalid:
            self.errors.report(
                'Invalid types provided as link "{}" options: {}'
                .format(self._format_path(obj), self._format_types(invalid))
            )
            return

        with self.push_ctx(obj):
            super(GraphValidator, self).visit_link(obj)

        graph_nodes_map = {e.name for e in self.items if e.name is not None}
        if obj.node not in graph_nodes_map:
            self.errors.report(
                'Link "{}" points to the missing node "{}"'
                .format(self._format_path(obj), obj.node)
            )

        if obj.requires is not None:
            if obj.requires not in self.ctx.fields_map:
                self.errors.report(
                    'Link "{}" requires missing field "{}" in the "{}" node'
                    .format(obj.name, obj.requires, self._format_path())
                )

        self._validate_deprecated_duplicates(obj)

    def visit_node(self, obj):
        node_name = obj.name or 'root'
        invalid = [f for f in obj.fields
                   if not isinstance(f, self._node_accept_types)]
        if invalid:
            self.errors.report(
                'Node can not contain these types: {} in node "{}"'
                .format(self._format_types(invalid), node_name)
            )
            return

        with self.push_ctx(obj):
            for item in obj.fields:
                self.visit(item)

        duplicates = self._get_duplicates(e.name for e in obj.fields)
        if duplicates:
            self.errors.report(
                'Duplicated names found in the "{}" node: {}'
                .format(node_name, self._format_names(duplicates))
            )

        if sum((1 for d in obj.directives if isinstance(d, Deprecated))) > 0:
            self.errors.report('Deprecated directive can not be used in Node')

    def visit_root(self, obj):
        self.visit_node(obj)

    def visit_graph(self, obj):
        self.visit_graph_items(obj.items)

    def visit_graph_items(self, items):
        invalid = [f for f in items
                   if not isinstance(f, self._graph_accept_types)]
        if invalid:
            self.errors.report(
                'Graph can not contain these types: {}'
                .format(self._format_types(invalid))
            )
            return

        root = Root(list(chain.from_iterable(e.fields for e in items
                                             if e.name is None)))
        self.visit(root)

        for item in items:
            if item.name is not None:
                self.visit(item)

        duplicates = self._get_duplicates(e.name for e in items
                                          if e.name is not None)
        if duplicates:
            self.errors.report('Duplicated nodes found in the graph: {}'
                               .format(self._format_names(duplicates)))



import inspect
import warnings
import dataclasses

from typing import Any
from functools import partial
from itertools import chain, repeat
from collections import defaultdict
from collections.abc import Sequence, Mapping, Hashable

from . import query as hiku_query
from .graph import Link, Maybe, One, Many, Nothing, Field
from .result import Proxy, Index, ROOT, Reference
from .executors.queue import Workflow, Queue


def _yield_options(graph_obj, query_obj):
    options = query_obj.options or {}
    for option in graph_obj.options:
        value = options.get(option.name, option.default)
        if value is Nothing:
            raise TypeError('Required option "{}" for {!r} was not provided'
                            .format(option.name, graph_obj))
        else:
            yield option.name, value


def _get_options(graph_obj, query_obj):
    return dict(_yield_options(graph_obj, query_obj))


class InitOptions(hiku_query.QueryTransformer):

    def __init__(self, graph):
        self._graph = graph
        self._path = [graph.root]

    def visit_field(self, obj):
        graph_obj = self._path[-1].fields_map[obj.name]
        if graph_obj.options:
            return obj.copy(options=_get_options(graph_obj, obj))
        else:
            return obj

    def visit_link(self, obj):
        graph_obj = self._path[-1].fields_map[obj.name]

        if isinstance(graph_obj, Link):
            self._path.append(self._graph.nodes_map[graph_obj.node])
            try:
                node = self.visit(obj.node)
            finally:
                self._path.pop()
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            node = obj.node

        options = _get_options(graph_obj, obj) if graph_obj.options else None
        return obj.copy(node=node, options=options)


class SplitQuery(hiku_query.QueryVisitor):

    def __init__(self, graph_node):
        self._node = graph_node
        self._fields = []
        self._links = []

    def split(self, query_node):
        for item in query_node.fields:
            self.visit(item)
        return self._fields, self._links

    def visit_node(self, obj):
        raise ValueError('Unexpected value: {!r}'.format(obj))

    def visit_field(self, obj):
        graph_obj = self._node.fields_map[obj.name]
        func = getattr(graph_obj.func, '__subquery__', graph_obj.func)
        self._fields.append((func, graph_obj, obj))

    def visit_link(self, obj):
        graph_obj = self._node.fields_map[obj.name]
        if isinstance(graph_obj, Link):
            if graph_obj.requires:
                self.visit(hiku_query.Field(graph_obj.requires))
            self._links.append((graph_obj, obj))
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            # `obj` here is a link, but this link is treated as a complex field
            func = getattr(graph_obj.func, '__subquery__', graph_obj.func)
            self._fields.append((func, graph_obj, obj))


class GroupQuery(hiku_query.QueryVisitor):

    def __init__(self, node):
        self._node = node
        self._funcs = []
        self._groups = []
        self._current_func = None

    def group(self, node):
        for item in node.fields:
            self.visit(item)
        return list(zip(self._funcs, self._groups))

    def visit_node(self, obj):
        raise ValueError('Unexpected value: {!r}'.format(obj))

    def visit_field(self, obj):
        graph_obj = self._node.fields_map[obj.name]
        func = getattr(graph_obj.func, '__subquery__', graph_obj.func)
        if func == self._current_func:
            self._groups[-1].append((graph_obj, obj))
        else:
            self._groups.append([(graph_obj, obj)])
            self._funcs.append(func)
            self._current_func = func

    def visit_link(self, obj):
        graph_obj = self._node.fields_map[obj.name]
        if graph_obj.requires:
            self.visit(hiku_query.Field(graph_obj.requires))
        self._groups.append((graph_obj, obj))
        self._funcs.append(graph_obj.func)
        self._current_func = None


def _check_store_fields(node, fields, ids, result):
    if node.name is not None:
        assert ids is not None
        if (
            isinstance(result, Sequence)
            and len(result) == len(ids)
            and all(isinstance(r, Sequence) and len(r) == len(fields)
                    for r in result)
        ):
            return
        else:
            expected = ('list (len: {}) of lists (len: {})'
                        .format(len(ids), len(fields)))
    else:
        if isinstance(result, Sequence) and len(result) == len(fields):
            return
        else:
            expected = 'list (len: {})'.format(len(fields))
    raise TypeError('Can\'t store field values, node: {!r}, fields: {!r}, '
                    'expected: {}, returned: {!r}'
                    .format(node.name or '__root__', [f.name for f in fields],
                            expected, result))


def _is_hashable(obj: Any) -> bool:
    if not isinstance(obj, Hashable):
        return False

    try:
        hash(obj)
    except TypeError:
        return False

    return True


def store_fields(index, node, query_fields, ids, query_result):
    if inspect.isgenerator(query_result):
        warnings.warn('Data loading functions should not return generators',
                      DeprecationWarning)
        query_result = list(query_result)

    _check_store_fields(node, query_fields, ids, query_result)

    names = [f.index_key for f in query_fields]
    if node.name is not None:
        assert ids is not None
        node_idx = index[node.name]
        for i, row in zip(ids, query_result):
            node_idx[i].update(zip(names, row))
    else:
        assert ids is None
        index.root.update(zip(names, query_result))


def link_reqs(index, node, link, ids):
    if node.name is not None:
        assert ids is not None
        node_idx = index[node.name]
        return [node_idx[i][link.requires] for i in ids]
    else:
        return index.root[link.requires]


def link_ref_maybe(graph_link, ident):
    if ident is Nothing:
        return None
    else:
        return Reference(graph_link.node, ident)


def link_ref_one(graph_link, ident):
    assert ident is not Nothing
    return Reference(graph_link.node, ident)


def link_ref_many(graph_link, idents):
    return [Reference(graph_link.node, i) for i in idents]


_LINK_REF_MAKER = {
    Maybe: link_ref_maybe,
    One: link_ref_one,
    Many: link_ref_many,
}


HASH_HINT = "\nHint: Consider adding __hash__ method or use hashable type."
DATACLASS_HINT = "\nHint: Use @dataclass(frozen=True) to make object hashable."


def _hashable_hint(obj: Any) -> str:
    if isinstance(obj, Sequence):
        return _hashable_hint(obj[0])

    if (
        dataclasses.is_dataclass(obj)
        and not getattr(obj, dataclasses._PARAMS).frozen  # type: ignore[attr-defined]  # noqa: E501
    ):
        return DATACLASS_HINT

    return HASH_HINT


def _check_store_links(node, link, ids, result):
    hint = ''
    if node.name is not None and link.requires is not None:
        assert ids is not None
        if link.type_enum is Maybe or link.type_enum is One:
            if isinstance(result, Sequence) and len(result) == len(ids):
                if all(map(_is_hashable, result)):
                    return
                expected = 'list of hashable objects'
                hint = _hashable_hint(result)
            else:
                expected = 'list (len: {})'.format(len(ids))
        elif link.type_enum is Many:
            if (
                isinstance(result, Sequence)
                and len(result) == len(ids)
                and all(isinstance(r, Sequence) for r in result)
            ):
                unhashable = False
                for items in result:
                    if not all(map(_is_hashable, items)):
                        unhashable = True
                        break

                if not unhashable:
                    return
                expected = 'list (len: {}) of lists of hashable objects'.format(len(ids))  # noqa: E501
                hint = _hashable_hint(result)
            else:
                expected = 'list (len: {}) of lists'.format(len(ids))
        else:
            raise TypeError(link.type_enum)
    else:
        if link.type_enum is Maybe or link.type_enum is One:
            if _is_hashable(result):
                return
            expected = 'hashable object'
            hint = _hashable_hint(result)
        elif link.type_enum is Many:
            if isinstance(result, Sequence):
                if all(map(_is_hashable, result)):
                    return
                expected = 'list of hashable objects'
                hint = _hashable_hint(result)
            else:
                expected = 'list'
        else:
            raise TypeError(link.type_enum)
    raise TypeError('Can\'t store link values, node: {!r}, link: {!r}, '
                    'expected: {}, returned: {!r}{}'
                    .format(node.name or '__root__', link.name,
                            expected, result, hint))


def store_links(index, node, graph_link, query_link, ids, query_result):
    _check_store_links(node, graph_link, ids, query_result)

    field_val = partial(_LINK_REF_MAKER[graph_link.type_enum], graph_link)
    if node.name is not None:
        assert ids is not None
        if graph_link.requires is None:
            query_result = repeat(query_result, len(ids))

        node_idx = index[node.name]
        for i, res in zip(ids, query_result):
            node_idx[i][query_link.index_key] = field_val(res)
    else:
        index.root[query_link.index_key] = field_val(query_result)


def link_result_to_ids(from_list, link_type, result):
    if from_list:
        if link_type is Maybe:
            return [i for i in result if i is not Nothing]
        elif link_type is One:
            if any(i is Nothing for i in result):
                raise TypeError('Non-optional link should not return Nothing: '
                                '{!r}'.format(result))
            return result
        elif link_type is Many:
            return list(chain.from_iterable(result))
    else:
        if link_type is Maybe:
            return [] if result is Nothing else [result]
        elif link_type is One:
            if result is Nothing:
                raise TypeError('Non-optional link should not return Nothing')
            return [result]
        elif link_type is Many:
            return result
    raise TypeError(repr([from_list, link_type]))


class Query(Workflow):

    def __init__(self, queue, task_set, graph, query, ctx):
        self._queue = queue
        self._task_set = task_set
        self._graph = graph
        self._query = query
        self._ctx = ctx
        self._index = Index()

    def _submit(self, func, *args, **kwargs):
        if _do_pass_context(func):
            return self._task_set.submit(func, self._ctx, *args, **kwargs)
        else:
            return self._task_set.submit(func, *args, **kwargs)

    def start(self):
        self.process_node(self._graph.root, self._query, None)

    def result(self):
        self._index.finish()
        return Proxy(self._index, ROOT, self._query)

    def _process_node_ordered(self, node, query, ids):
        proc_steps = GroupQuery(node).group(query)

        # recursively and sequentially schedule fields and links
        def proc(steps):
            step_func, step_item = steps.pop(0)
            if isinstance(step_item, list):
                dep = self._schedule_fields(node, step_func, step_item, ids)
            else:
                graph_link, query_link = step_item
                dep = self._schedule_link(node, graph_link, query_link, ids)

            if steps:
                self._queue.add_callback(dep, lambda: proc(steps))

        if proc_steps:
            proc(proc_steps)

    def process_node(self, node, query, ids):
        if query.ordered:
            self._process_node_ordered(node, query, ids)
            return

        fields, links = SplitQuery(node).split(query)

        to_func = {}
        from_func = defaultdict(list)
        for func, graph_field, query_field in fields:
            to_func[graph_field.name] = func
            from_func[func].append((graph_field, query_field))

        # schedule fields resolve
        to_dep = {}
        for func, func_fields in from_func.items():
            to_dep[func] = self._schedule_fields(node, func, func_fields, ids)

        # schedule link resolve
        for graph_link, query_link in links:
            schedule = partial(self._schedule_link, node,
                               graph_link, query_link, ids)
            if graph_link.requires:
                dep = to_dep[to_func[graph_link.requires]]
                self._queue.add_callback(dep, schedule)
            else:
                schedule()

    def process_link(self, node, graph_link, query_link, ids, result):
        if inspect.isgenerator(result):
            warnings.warn('Data loading functions should not return generators',
                          DeprecationWarning)
            result = list(result)
        store_links(self._index, node, graph_link, query_link, ids, result)
        from_list = ids is not None and graph_link.requires is not None
        to_ids = link_result_to_ids(from_list, graph_link.type_enum, result)
        if to_ids:
            self.process_node(self._graph.nodes_map[graph_link.node],
                              query_link.node, to_ids)

    def _schedule_fields(self, node, func, fields, ids):
        query_fields = [qf for _, qf in fields]
        if hasattr(func, '__subquery__'):
            assert ids is not None
            dep = self._queue.fork(self._task_set)
            proc = func(fields, ids, self._queue, self._ctx, dep)
        else:
            if ids is None:
                dep = self._submit(func, query_fields)
            else:
                dep = self._submit(func, query_fields, ids)
            proc = dep.result
        self._queue.add_callback(
            dep,
            lambda: store_fields(self._index, node, query_fields, ids, proc())
        )
        return dep

    def _schedule_link(self, node, graph_link, query_link, ids):
        args = []
        if graph_link.requires:
            args.append(link_reqs(self._index, node, graph_link, ids))
        if graph_link.options:
            args.append(query_link.options)
        dep = self._submit(graph_link.func, *args)
        self._queue.add_callback(dep, (
            lambda:
            self.process_link(node, graph_link, query_link, ids, dep.result())
        ))
        return dep


def pass_context(func):
    func.__pass_context__ = True
    return func


def _do_pass_context(func):
    return getattr(func, '__pass_context__', False)


class Context(Mapping):

    def __init__(self, mapping):
        self.__mapping = mapping

    def __len__(self):
        return len(self.__mapping)

    def __iter__(self):
        return iter(self.__mapping)

    def __getitem__(self, item):
        try:
            return self.__mapping[item]
        except KeyError:
            raise KeyError('Key {!r} is not specified '
                           'in the query context'.format(item))


class Engine:

    def __init__(self, executor):
        self.executor = executor

    def execute(self, graph, query, ctx=None):
        if ctx is None:
            ctx = {}
        query = InitOptions(graph).visit(query)
        queue = Queue(self.executor)
        task_set = queue.fork(None)
        query_workflow = Query(queue, task_set, graph, query, Context(ctx))
        query_workflow.start()
        return self.executor.process(queue, query_workflow)



from typing import List

from ..graph import Graph

from .directive import Key


def get_keys(graph: Graph, typename: str) -> List[int]:
    """Get all 'key' directives fields"""
    node = graph.nodes_map[typename]
    return [d.fields for d in node.directives if isinstance(d, Key)]



"""
Based on the code from https://github.com/gns24/pydatomic project
"""
from uuid import UUID
from decimal import Decimal
from datetime import datetime
from itertools import chain
from json.encoder import encode_basestring, encode_basestring_ascii  # type: ignore # noqa: E501


class ImmutableDict(dict):
    _hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(frozenset(self.items()))
        return self._hash

    def _immutable(self):
        raise TypeError("{} object is immutable"
                        .format(self.__class__.__name__))

    __delitem__ = __setitem__ = _immutable  # type: ignore
    clear = pop = popitem = setdefault = update = _immutable  # type: ignore


class Symbol(str):

    def __repr__(self):
        return self

    def __eq__(self, other):
        return isinstance(other, type(self)) and \
            super(Symbol, self).__eq__(other)

    def __hash__(self):
        return super(Symbol, self).__hash__()


class Keyword(str):

    def __repr__(self):
        return ':{}'.format(self)

    def __eq__(self, other):
        return isinstance(other, type(self)) and \
            super(Keyword, self).__eq__(other)

    def __hash__(self):
        return super(Keyword, self).__hash__()


class List(tuple):

    def __repr__(self):
        return '[{}]'.format(' '.join(map(repr, self)))


class Tuple(tuple):

    def __repr__(self):
        return '({})'.format(' '.join(map(repr, self)))


class Dict(ImmutableDict):

    def __repr__(self):
        return '{{{}}}'.format(' '.join('{!r} {!r}'.format(*i)
                               for i in self.items()))


class Set(frozenset):

    def __repr__(self):
        return '#{{{}}}'.format(' '.join(map(repr, self)))


class TaggedElement:

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return '#{} {!r}'.format(self.name, self.value)

    def __eq__(self, other):
        return isinstance(other, type(self)) and \
            self.name == other.name and self.value == other.value


def coroutine(func):
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr
    return start


@coroutine
def appender(lst):
    while True:
        lst.append((yield))


def inst_handler(time_string):
    return datetime.strptime(time_string, '%Y-%m-%dT%H:%M:%S.%fZ')


TAG_HANDLERS = {'inst': inst_handler, 'uuid': UUID}

STOP_CHARS = " ,\n\r\t"

_CHAR_HANDLERS = {
    'newline': '\n',
    'space': ' ',
    'tab': '\t',
}

_CHAR_MAP = {
    "a": "\a",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
    "v": "\v",
}

_END_CHARS = {
    '#': '}',
    '{': '}',
    '[': ']',
    '(': ')',
}

_NIL = object()


@coroutine
def tag_handler(tag_name, tag_handlers):
    while True:
        c = (yield)
        if c in STOP_CHARS + '{"[(\\#':
            break
        tag_name += c
    elements = []
    handler = parser(appender(elements), tag_handlers)
    handler.send(c)
    while not elements:
        handler.send((yield))
    if tag_name in tag_handlers:
        yield tag_handlers[tag_name](elements[0]), True
    else:
        yield TaggedElement(tag_name, elements[0]), True
        yield None, True


@coroutine
def character_handler():
    r = (yield)
    while 1:
        c = (yield)
        if not c.isalpha():
            if len(r) == 1:
                yield r, False
            else:
                yield _CHAR_HANDLERS[r], False
        r += c


def parse_number(s):
    s = s.rstrip('MN').upper()
    if 'E' not in s and '.' not in s:
        return int(s)
    return float(s)


@coroutine
def number_handler(s):
    while 1:
        c = (yield)
        if c in "0123456789+-eEMN.":
            s += c
        else:
            yield parse_number(s), False


@coroutine
def symbol_handler(s):
    while 1:
        c = (yield)
        if c in '}])' + STOP_CHARS:
            if s[0] == ':':
                yield Keyword(s[1:]), False
            elif s == 'true':
                yield True, False
            elif s == 'false':
                yield False, False
            elif s == 'nil':
                yield _NIL, False
            else:
                yield Symbol(s), False
        else:
            s += c


@coroutine
def parser(target, tag_handlers, stop=None):
    handler = None
    while True:
        c = (yield)
        if handler:
            v = handler.send(c)
            if v is None:
                continue
            else:
                handler = None
                v, consumed = v
                if v is not None:
                    if v is _NIL:
                        target.send(None)
                    else:
                        target.send(v)
                if consumed:
                    continue
        if c == stop:
            return
        if c in STOP_CHARS:
            continue
        if c == ';':
            while (yield) != '\n':
                pass
        elif c == '"':
            chars = []
            while 1:
                char = (yield)
                if char == '\\':
                    char = (yield)
                    char2 = _CHAR_MAP.get(char)
                    if char2 is not None:
                        chars.append(char2)
                    else:
                        chars.append(char)
                elif char == '"':
                    target.send(''.join(chars))
                    break
                else:
                    chars.append(char)
        elif c == '\\':
            handler = character_handler()
        elif c in '0123456789':
            handler = number_handler(c)
        elif c in '-.':
            c2 = (yield)
            if c2.isdigit():  # .5 should be an error
                handler = number_handler(c + c2)
            else:
                handler = symbol_handler(c + c2)
        elif c.isalpha() or c == ':':
            handler = symbol_handler(c)
        elif c in '[({#':
            if c == '#':
                c2 = (yield)
                if c2 != '{':
                    handler = tag_handler(c2, tag_handlers)
                    continue
            end_char = _END_CHARS[c]
            lst = []
            p = parser(appender(lst), tag_handlers, stop=end_char)
            try:
                while 1:
                    p.send((yield))
            except StopIteration:
                pass
            if c == '[':
                target.send(List(lst))
            elif c == '(':
                target.send(Tuple(lst))
            elif c == '{':
                if len(lst) % 2:
                    raise Exception("Map literal must contain an even "
                                    "number of elements")
                target.send(Dict(zip(lst[::2], lst[1::2])))
            else:
                target.send(Set(lst))
        else:
            raise ValueError("Unexpected character in edn", c)


def loads(s, tag_handlers=None):
    if not isinstance(s, str):
        raise TypeError('The EDN value must be "str", not {!r}'
                        .format(type(s).__name__))
    lst = []
    target = parser(appender(lst), dict(tag_handlers or (), **TAG_HANDLERS))
    for c in s:
        target.send(c)
    target.send(' ')
    if len(lst) != 1:
        raise ValueError("Expected exactly one top-level element "
                         "in edn string", s)
    return lst[0]


def _iterencode_items(items, default, encoder):
    items_iter = iter(items)
    try:
        first = next(items_iter)
    except StopIteration:
        return
    for chunk in _iterencode(first, default, encoder):
        yield chunk
    while True:
        try:
            next_item = next(items_iter)
        except StopIteration:
            return
        yield ' '
        for chunk in _iterencode(next_item, default, encoder):
            yield chunk


def _default(obj):
    raise ValueError('{!r} is not EDN serializable'.format(obj))


def _iterencode(obj, default, encoder):
    if obj is None:
        yield 'nil'
    elif obj is True:
        yield 'true'
    elif obj is False:
        yield 'false'
    elif isinstance(obj, int):
        yield str(int(obj))
    elif isinstance(obj, float):
        # FIXME: proper float encoding
        yield str(float(obj))
    elif isinstance(obj, Decimal):
        yield '{}M'.format(obj)
    elif isinstance(obj, Keyword):
        yield ':{}'.format(obj)
    elif isinstance(obj, Symbol):
        yield obj
    elif isinstance(obj, str):
        yield encoder(obj)
    elif isinstance(obj, (list, List)):
        # NOTE: `(list, List)` check should be before `(tuple, Tuple)`,
        # because `List` is implemented as tuple subclass
        yield '['
        for chunk in _iterencode_items(obj, default, encoder):
            yield chunk
        yield ']'
    elif isinstance(obj, (tuple, Tuple)):
        yield '('
        for chunk in _iterencode_items(obj, default, encoder):
            yield chunk
        yield ')'
    elif isinstance(obj, (dict, Dict)):
        yield '{'
        for chunk in _iterencode_items(chain.from_iterable(obj.items()),
                                       default, encoder):
            yield chunk
        yield '}'
    elif isinstance(obj, (set, Set)):
        yield '#{'
        for chunk in _iterencode_items(obj, default, encoder):
            yield chunk
        yield '}'
    elif isinstance(obj, datetime):
        # FIXME: proper RFC-3339 encoding
        assert not obj.tzinfo
        yield obj.strftime('#inst "%Y-%m-%dT%H:%M:%S.%fZ"')
    elif isinstance(obj, UUID):
        yield '#uuid "{}"'.format(obj)
    elif isinstance(obj, TaggedElement):
        yield '#{} '.format(obj.name)
        for chunk in _iterencode(obj.value, _default, encoder):
            yield chunk
    else:
        obj = default(obj)
        for chunk in _iterencode(obj, default, encoder):
            yield chunk


def dumps(obj, default=None, ensure_ascii=True):
    if default is None:
        default = _default
    if ensure_ascii:
        encoder = encode_basestring_ascii
    else:
        encoder = encode_basestring
    return ''.join(_iterencode(obj, default, encoder))



from typing import (
    Optional,
    Union,
)

from hiku.graph import (
    Field,
    Link,
)


class DirectiveBase:
    pass


class Deprecated(DirectiveBase):
    """
    https://spec.graphql.org/June2018/#sec--deprecated
    """
    def __init__(self, reason: Optional[str] = None):
        self.reason = reason

    def accept(self, visitor):
        return visitor.visit_deprecated_directive(self)


def get_deprecated(field: Union[Field, Link]) -> Optional[Deprecated]:
    """Get deprecated directive"""
    return next(
        (d for d in field.directives if isinstance(d, Deprecated)),
        None
    )



"""
    hiku.query
    ~~~~~~~~~~

    `Hiku` doesn't rely on any specific query language, internally it uses
    generic query representation to describe result and it could be constructed
    by parsing different suitable query languages.

    However, `Hiku` provides one built-in way to describe result -- `edn`_ data
    structure -- `simple` queries, which are very similar to the `om.next`_
    queries, which are inspired by `Datomic Pull API`_.

      - ``[:foo]`` - node fields definition (`edn` keywords in vector)
      - ``{:bar [:baz]}`` - link definition (`edn` map with keyword and vector)
      - ``(:foo {:key val})`` - field or link options definition (field name or
        link name, wrapped with `edn` list with map of options as a second
        value)

    Example:

    .. code-block:: clojure

        [:foo {:bar [:baz]}]

    This query will be read internally as:

    .. code-block:: python

        Node([Field('foo'),
              Link('bar', Node([Field('baz')]))])

    And query result will look like this:

    .. code-block:: python

        {
            'foo': 1,
            'bar': {
                'baz': 2,
            },
        }

    .. _edn: https://github.com/edn-format/edn
    .. _Datomic Pull API: http://docs.datomic.com/pull.html
    .. _om.next: https://github.com/omcljs/om/wiki/Documentation-(om.next)
"""
import typing as t

from itertools import chain
from collections import OrderedDict
from collections.abc import Sequence

from .utils import cached_property


def _compute_hash(obj):
    if isinstance(obj, dict):
        return hash(tuple((_compute_hash(k), _compute_hash(v))
                          for k, v in sorted(obj.items())))
    elif isinstance(obj, list):
        return hash(tuple(_compute_hash(i) for i in obj))
    else:
        return hash(obj)


class Base:
    __attrs__: t.Tuple[str, ...] = ()

    def __repr__(self):
        kwargs = ', '.join('{}={!r}'.format(attr, self.__dict__[attr])
                           for attr in self.__attrs__)
        return '{}({})'.format(self.__class__.__name__, kwargs)

    def __eq__(self, other):
        return (self.__class__ is other.__class__
                and all(self.__dict__[attr] == other.__dict__[attr]
                        for attr in self.__attrs__))

    def __ne__(self, other):
        return not self.__eq__(other)

    def copy(self, **kwargs):
        obj = self.__class__.__new__(self.__class__)
        obj.__dict__.update((attr, kwargs.get(attr, self.__dict__[attr]))
                            for attr in self.__attrs__)
        return obj


class FieldBase(Base):

    @cached_property
    def result_key(self):
        if self.alias is not None:
            return self.alias
        else:
            return self.name

    @cached_property
    def options_hash(self):
        if self.options:
            return _compute_hash(self.options)
        else:
            return None

    @cached_property
    def index_key(self):
        if self.options_hash is not None:
            return '{}[{}]'.format(self.name, self.options_hash)
        else:
            return self.name


class Field(FieldBase):
    """Represents a field of the node

    :param name: name of the field
    :param optional options: field options -- mapping of names to values
    :param optional alias: field's name in result
    """
    __attrs__ = ('name', 'options', 'alias')

    def __init__(self, name, options=None, alias=None):
        self.name = name
        self.options = options
        self.alias = alias

    def accept(self, visitor):
        return visitor.visit_field(self)


class Link(FieldBase):
    """Represents a link to the node

    :param name: name of the link
    :param node: collection of fields and links --
                 :py:class:`~hiku.query.Node`
    :param optional options: link options -- mapping of names to values
    :param optional alias: link's name in result
    """
    __attrs__ = ('name', 'node', 'options', 'alias')

    def __init__(self, name, node, options=None, alias=None):
        self.name = name
        self.node = node
        self.options = options
        self.alias = alias

    def accept(self, visitor):
        return visitor.visit_link(self)


class Node(Base):
    """Represents collection of fields and links

    :param fields: list of :py:class:`~hiku.query.Field` and
        :py:class:`~hiku.query.Link`
    :param ordered: whether to compute fields of this node sequentially
        in order or not
    """
    __attrs__ = ('fields', 'ordered')

    def __init__(self, fields, ordered=False):
        self.fields = fields
        self.ordered = ordered

    @cached_property
    def fields_map(self):
        return OrderedDict((f.name, f) for f in self.fields)

    @cached_property
    def result_map(self):
        return OrderedDict((f.result_key, f) for f in self.fields)

    def accept(self, visitor):
        return visitor.visit_node(self)


def _merge(nodes):
    fields = set()
    links = {}
    to_merge = OrderedDict()
    for field in chain.from_iterable(e.fields for e in nodes):
        key = (field.name, field.options_hash, field.alias)
        if field.__class__ is Link:
            if key not in to_merge:
                to_merge[key] = [field.node]
                links[key] = field
            else:
                to_merge[key].append(field.node)
        else:
            if key not in fields:
                fields.add(key)
                yield field
    for key, values in to_merge.items():
        link = links[key]
        yield link.copy(node=merge(values))


def merge(nodes):
    """Merges multiple queries into one query

    :param nodes: queries, represented as list of :py:class:`~hiku.query.Node`
    :return: merged query as one :py:class:`~hiku.query.Node`
    """
    assert isinstance(nodes, Sequence), type(nodes)
    ordered = any(n.ordered for n in nodes)
    return Node(list(_merge(nodes)), ordered=ordered)


class QueryVisitor:

    def visit(self, obj):
        return obj.accept(self)

    def visit_field(self, obj):
        pass

    def visit_link(self, obj):
        self.visit(obj.node)

    def visit_node(self, obj):
        for item in obj.fields:
            self.visit(item)


class QueryTransformer:

    def visit(self, obj):
        return obj.accept(self)

    def visit_field(self, obj):
        return obj.copy()

    def visit_link(self, obj):
        return obj.copy(node=self.visit(obj.node))

    def visit_node(self, obj):
        return obj.copy(fields=[self.visit(f) for f in obj.fields])



from collections import deque

from hiku.denormalize.graphql import DenormalizeGraphQL


class DenormalizeEntityGraphQL(DenormalizeGraphQL):
    def __init__(self, graph, result, root_type_name):
        super().__init__(graph, result, root_type_name)
        self._type = deque([graph.__types__[root_type_name]])



from collections import namedtuple


def _namedtuple(typename, field_names):
    """Fixes hashing for different tuples with same content
    """
    base = namedtuple(typename, field_names)

    def __hash__(self):
        return hash((self.__class__, super(base, self).__hash__()))

    return type(typename, (base,), {
        '__slots__': (),
        '__hash__': __hash__,
    })


SCALAR = _namedtuple('SCALAR', 'name')
OBJECT = _namedtuple('OBJECT', 'name')
DIRECTIVE = _namedtuple('DIRECTIVE', 'name')
INPUT_OBJECT = _namedtuple('INPUT_OBJECT', 'name')
LIST = _namedtuple('LIST', 'of_type')
NON_NULL = _namedtuple('NON_NULL', 'of_type')

FieldIdent = _namedtuple('FieldIdent', 'node, name')
FieldArgIdent = _namedtuple('FieldArgIdent', 'node, field, name')
InputObjectFieldIdent = _namedtuple('InputObjectFieldIdent', 'name, key')
DirectiveArgIdent = _namedtuple('DirectiveArgIdent', 'name, arg')



import enum

from graphql.language import ast
from graphql.language.parser import parse

from ..query import Node, Field, Link, merge


class NodeVisitor:

    def visit(self, obj):
        visit_method = getattr(self, 'visit_{}'.format(obj.kind))
        if visit_method is None:
            raise NotImplementedError('Not implemented node type: {!r}'
                                      .format(obj))
        return visit_method(obj)

    def visit_document(self, obj):
        for definition in obj.definitions:
            self.visit(definition)

    def visit_operation_definition(self, obj):
        self.visit(obj.selection_set)

    def visit_fragment_definition(self, obj):
        self.visit(obj.selection_set)

    def visit_selection_set(self, obj):
        for i in obj.selections:
            self.visit(i)

    def visit_field(self, obj):
        pass

    def visit_fragment_spread(self, obj):
        pass

    def visit_inline_fragment(self, obj):
        self.visit(obj.selection_set)


class OperationGetter(NodeVisitor):

    def __init__(self, operation_name=None):
        self._operations = {}
        self._operation_name = operation_name

    @classmethod
    def get(cls, doc, operation_name=None):
        self = cls(operation_name=operation_name)
        self.visit(doc)
        if not self._operations:
            raise TypeError('No operations in the document')

        if self._operation_name is None:
            if len(self._operations) > 1:
                raise TypeError('Document should contain exactly one operation '
                                'when no operation name was provided')
            return next(iter(self._operations.values()))
        else:
            try:
                return self._operations[self._operation_name]
            except KeyError:
                raise ValueError('Undefined operation name: {!r}'
                                 .format(self._operation_name))

    def visit_fragment_definition(self, obj):
        pass  # skip visit here

    def visit_operation_definition(self, obj):
        name = obj.name.value if obj.name is not None else None
        if name in self._operations:
            raise TypeError('Duplicate operation definition: {!r}'
                            .format(name))
        self._operations[name] = obj


class FragmentsCollector(NodeVisitor):

    def __init__(self):
        self.fragments_map = {}

    def visit_operation_definition(self, obj):
        pass  # not interested in operations here

    def visit_fragment_definition(self, obj):
        if obj.name.value in self.fragments_map:
            raise TypeError('Duplicated fragment name: "{}"'
                            .format(obj.name.value))
        self.fragments_map[obj.name.value] = obj


class SelectionSetVisitMixin:

    def transform_fragment(self, name):
        raise NotImplementedError(type(self))

    @property
    def query_variables(self):
        raise NotImplementedError(type(self))

    @property
    def query_name(self):
        raise NotImplementedError(type(self))

    def lookup_variable(self, name):
        try:
            return self.query_variables[name]
        except KeyError:
            raise TypeError('Variable ${} is not defined in query {}'
                            .format(name, self.query_name or '<unnamed>'))

    def visit_selection_set(self, obj):
        for i in obj.selections:
            for j in self.visit(i):
                yield j

    def _should_skip(self, obj):
        if not obj.directives:
            return

        skip = next((d for d in obj.directives if d.name.value == 'skip'),
                    None)
        if skip is not None:
            if len(skip.arguments) != 1:
                raise TypeError('@skip directive accepts exactly one '
                                'argument, {} provided'
                                .format(len(skip.arguments)))
            skip_arg = skip.arguments[0]
            if skip_arg.name.value != 'if':
                raise TypeError('@skip directive does not accept "{}" '
                                'argument'
                                .format(skip_arg.name.value))
            return self.visit(skip_arg.value)

        include = next((d for d in obj.directives if d.name.value == 'include'),
                       None)
        if include is not None:
            if len(include.arguments) != 1:
                raise TypeError('@include directive accepts exactly one '
                                'argument, {} provided'
                                .format(len(include.arguments)))
            include_arg = include.arguments[0]
            if include_arg.name.value != 'if':
                raise TypeError('@include directive does not accept "{}" '
                                'argument'
                                .format(include_arg.name.value))
            return not self.visit(include_arg.value)

    def visit_field(self, obj):
        if self._should_skip(obj):
            return

        if obj.arguments:
            options = {arg.name.value: self.visit(arg.value)
                       for arg in obj.arguments}
        else:
            options = None

        if obj.alias is not None:
            alias = obj.alias.value
        else:
            alias = None

        if obj.selection_set is None:
            yield Field(obj.name.value, options=options, alias=alias)
        else:
            node = Node(list(self.visit(obj.selection_set)))
            yield Link(obj.name.value, node, options=options, alias=alias)

    def visit_variable(self, obj):
        return self.lookup_variable(obj.name.value)

    def visit_null_value(self, obj):
        return None

    def visit_int_value(self, obj):
        return int(obj.value)

    def visit_float_value(self, obj):
        return float(obj.value)

    def visit_string_value(self, obj):
        return obj.value

    def visit_boolean_value(self, obj):
        return obj.value

    def visit_enum_value(self, obj):
        return obj.value

    def visit_list_value(self, obj):
        return [self.visit(i) for i in obj.values]

    def visit_object_value(self, obj):
        return {f.name.value: self.visit(f.value) for f in obj.fields}

    def visit_fragment_spread(self, obj):
        if self._should_skip(obj):
            return
        for i in self.transform_fragment(obj.name.value):
            yield i

    def visit_inline_fragment(self, obj):
        if self._should_skip(obj):
            return
        for i in self.visit(obj.selection_set):
            yield i


class FragmentsTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name = None
    query_variables = None

    def __init__(self, document, query_name, query_variables):
        collector = FragmentsCollector()
        collector.visit(document)
        self.query_name = query_name
        self.query_variables = query_variables
        self.fragments_map = collector.fragments_map
        self.cache = {}
        self.pending_fragments = set()

    def transform_fragment(self, name):
        return self.visit(self.fragments_map[name])

    def visit_operation_definition(self, obj):
        pass  # not interested in operations here

    def visit_fragment_definition(self, obj):
        if obj.name.value in self.cache:
            return self.cache[obj.name.value]
        else:
            if obj.name.value in self.pending_fragments:
                raise TypeError('Cyclic fragment usage: "{}"'
                                .format(obj.name.value))
            self.pending_fragments.add(obj.name.value)
            try:
                selection_set = list(self.visit(obj.selection_set))
            finally:
                self.pending_fragments.discard(obj.name.value)
            self.cache[obj.name.value] = selection_set
            return selection_set


class GraphQLTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name = None
    query_variables = None
    fragments_transformer = None

    def __init__(self, document, variables=None):
        self.document = document
        self.variables = variables

    @classmethod
    def transform(cls, document, op, variables=None):
        visitor = cls(document, variables)
        return visitor.visit(op)

    def transform_fragment(self, name):
        return self.fragments_transformer.transform_fragment(name)

    def visit_operation_definition(self, obj):
        variables = self.variables or {}
        query_name = obj.name.value if obj.name else '<unnamed>'
        query_variables = {}
        for var_defn in obj.variable_definitions or ():
            name = var_defn.variable.name.value
            try:
                value = variables[name]  # TODO: check variable type
            except KeyError:
                if var_defn.default_value is not None:
                    value = self.visit(var_defn.default_value)
                elif isinstance(var_defn.type, ast.NonNullTypeNode):
                    raise TypeError('Variable "{}" is not provided for query {}'
                                    .format(name, query_name))
                else:
                    value = None
            query_variables[name] = value

        self.query_name = query_name
        self.query_variables = query_variables
        self.fragments_transformer = FragmentsTransformer(self.document,
                                                          self.query_name,
                                                          self.query_variables)
        ordered = obj.operation is ast.OperationType.MUTATION
        try:
            node = Node(list(self.visit(obj.selection_set)),
                        ordered=ordered)
        finally:
            self.query_name = None
            self.query_variables = None
            self.fragments_transformer = None
        return merge([node])


def read(src, variables=None, operation_name=None):
    """Reads a query from the GraphQL document

    Example:

    .. code-block:: python

        query = read('{ foo bar }')
        result = engine.execute(graph, query)

    :param str src: GraphQL query
    :param dict variables: query variables
    :param str operation_name: Name of the operation to execute
    :return: :py:class:`hiku.query.Node`, ready to execute query object
    """
    doc = parse(src)
    op = OperationGetter.get(doc, operation_name=operation_name)
    if op.operation is not ast.OperationType.QUERY:
        raise TypeError('Only "query" operations are supported, '
                        '"{}" operation was provided'
                        .format(op.operation.value))

    return GraphQLTransformer.transform(doc, op, variables)


class OperationType(enum.Enum):
    """Enumerates GraphQL operation types"""
    #: query operation
    QUERY = ast.OperationType.QUERY
    #: mutation operation
    MUTATION = ast.OperationType.MUTATION
    #: subscription operation
    SUBSCRIPTION = ast.OperationType.SUBSCRIPTION


class Operation:
    """Represents requested GraphQL operation"""
    def __init__(self, type_, query, name=None):
        #: type of the operation
        self.type = type_
        #: operation's query
        self.query = query
        #: optional name of the operation
        self.name = name


def read_operation(src, variables=None, operation_name=None):
    """Reads an operation from the GraphQL document

    Example:

    .. code-block:: python

        op = read_operation('{ foo bar }')
        if op.type is OperationType.QUERY:
            result = engine.execute(query_graph, op.query)

    :param str src: GraphQL document
    :param dict variables: query variables
    :param str operation_name: Name of the operation to execute
    :return: :py:class:`Operation`
    """
    doc = parse(src)
    op = OperationGetter.get(doc, operation_name=operation_name)
    query = GraphQLTransformer.transform(doc, op, variables)
    type_ = OperationType._value2member_map_.get(op.operation)
    name = op.name.value if op.name else None
    if type_ is None:
        raise TypeError('Unsupported operation type: {}'.format(op.operation))
    else:
        return Operation(type_, query, name)



from collections import deque

from ..query import Field, Link
from ..types import TypeRefMeta, SequenceMeta, OptionalMeta

from .base import Denormalize


class DenormalizeGraphQL(Denormalize):

    def __init__(self, graph, result, root_type_name):
        super().__init__(graph, result)
        self._type_name = deque([root_type_name])

    def visit_field(self, obj: Field):
        if obj.name == '__typename':
            self._res[-1][obj.result_key] = self._type_name[-1]
        else:
            super().visit_field(obj)

    def visit_link(self, obj: Link):
        type_ = self._type[-1].__field_types__[obj.name]
        if isinstance(type_, TypeRefMeta):
            type_ref = type_
        elif isinstance(type_, SequenceMeta):
            type_ref = type_.__item_type__
            assert isinstance(type_ref, TypeRefMeta), type_ref
        elif isinstance(type_, OptionalMeta):
            type_ref = type_.__type__
            assert isinstance(type_ref, TypeRefMeta), type_ref
        else:
            raise AssertionError(repr(type_))
        self._type_name.append(type_ref.__type_name__)
        super().visit_link(obj)
        self._type_name.pop()



from typing import List

from hiku.graph import Graph
from hiku.query import Node
from hiku.validate.query import validate as validate_query


def validate(graph: Graph, query: Node) -> List[str]:
    if (
        '_entities' in query.fields_map or
        '_service' in query.fields_map
    ):
        return []

    return validate_query(graph, query)



import inspect
from collections import defaultdict

from .sdl import print_sdl
from .utils import get_keys
from hiku.engine import (
    InitOptions,
    Query,
    Context,
)
from hiku.executors.queue import Queue
from hiku.graph import Graph
from hiku.result import (
    Proxy,
    Index,
    ROOT,
)
from ..query import (
    Node,
    Field,
)


class Engine:
    def __init__(self, executor):
        self.executor = executor

    def execute_service(self, graph):
        idx = Index()
        idx[ROOT.node] = Index()
        idx[ROOT.node][ROOT.ident] = {'sdl': print_sdl(graph)}
        return Proxy(idx, ROOT, Node(fields=[Field('sdl')]))

    def execute_entities(self, graph, query, ctx):
        entities_link = query.fields_map['_entities']
        query = entities_link.node
        representations = entities_link.options['representations']

        queue = Queue(self.executor)
        task_set = queue.fork(None)
        query_workflow = Query(queue, task_set, graph, query, Context(ctx))

        type_ids_map = defaultdict(list)

        for rep in representations:
            typename = rep['__typename']
            for key in get_keys(graph, typename):
                if key not in rep:
                    continue
                ident = rep[key]

                type_ids_map[typename].append(ident)

        for typename in type_ids_map:
            ids = type_ids_map[typename]
            node = graph.nodes_map[typename]
            query_workflow.process_node(node, query, ids)

        return self.executor.process(queue, query_workflow)

    def execute_query(self, graph, query, ctx):
        query = InitOptions(graph).visit(query)
        queue = Queue(self.executor)
        task_set = queue.fork(None)
        query_workflow = Query(queue, task_set, graph, query, Context(ctx))

        query_workflow.start()
        return self.executor.process(queue, query_workflow)

    def execute(self, graph: Graph, query, ctx=None) -> Proxy:
        if ctx is None:
            ctx = {}

        if '_service' in query.fields_map:
            return self.execute_service(graph)
        elif '_entities' in query.fields_map:
            return self.execute_entities(graph, query, ctx)

        return self.execute_query(graph, query, ctx)

    async def execute_async(self, graph: Graph, query, ctx=None) -> Proxy:
        result = self.execute(graph, query, ctx)
        if inspect.iscoroutine(result):
            return await result
        return result



from hiku.directives import DirectiveBase


class Key(DirectiveBase):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#key
    """
    def __init__(self, fields):
        self.fields = fields

    def accept(self, visitor):
        return visitor.visit_key_directive(self)


class Provides(DirectiveBase):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#provides
    """
    def __init__(self, fields):
        self.fields = fields

    def accept(self, visitor):
        return visitor.visit_provides_directive(self)


class Requires(DirectiveBase):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#requires
    """
    def __init__(self, fields):
        self.fields = fields

    def accept(self, visitor):
        return visitor.visit_requires_directive(self)


class External(DirectiveBase):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#external
    """
    def accept(self, visitor):
        return visitor.visit_external_directive(self)


class Extends(DirectiveBase):
    """
    Apollo Federation supports using an @extends directive in place of extend
    type to annotate type references
    https://www.apollographql.com/docs/federation/federation-spec/
    """
    def accept(self, visitor):
        return visitor.visit_extends_directive(self)



from contextlib import contextmanager
from collections import abc as collections_abc

from ..types import AbstractTypeVisitor
from ..query import QueryVisitor
from ..graph import GraphVisitor, Field, Link, Nothing

from .errors import Errors


_undefined = object()


class _AssumeRecord(AbstractTypeVisitor):

    def __init__(self, data_types, _nested=False):
        self._data_types = data_types
        self._nested = _nested

    def _get_nested(self):
        return _AssumeRecord(self._data_types, _nested=True)

    def visit(self, obj):
        if obj is not None:
            return obj.accept(self)

    def _false(self, obj):
        pass

    visit_any = _false
    visit_boolean = _false
    visit_string = _false
    visit_integer = _false
    visit_float = _false
    visit_mapping = _false
    visit_callable = _false

    def visit_optional(self, obj):
        if not self._nested:
            return self._get_nested().visit(obj.__type__)

    def visit_sequence(self, obj):
        if not self._nested:
            return self._get_nested().visit(obj.__item_type__)

    def visit_record(self, obj):
        # return fields alongside type definitions
        return obj.__field_types__

    def visit_typeref(self, obj):
        return self.visit(self._data_types[obj.__type_name__])


class _AssumeField(GraphVisitor):

    def __init__(self, node, errors):
        self.node = node
        self.errors = errors

    def visit_field(self, obj):
        return True

    def visit_link(self, obj):
        self.errors.report('Trying to query "{}.{}" link as it was a field'
                           .format(self.node.name or 'root', obj.name))
        return False

    def visit_node(self, obj):
        assert self.node.name is None,\
            'Nested node can be only in the root node'
        self.errors.report('Trying to query "{}" node as it was a field'
                           .format(obj.name))
        return False

    def visit_root(self, obj):
        raise AssertionError('Root node is not expected here')


class _OptionError(TypeError):

    def __init__(self, description):
        self.description = description
        super(_OptionError, self).__init__(description)


class _OptionTypeError(_OptionError):

    def __init__(self, value, expected):
        description = '"{}" instead of {!r}'.format(type(value).__name__,
                                                    expected)
        super(_OptionTypeError, self).__init__(description)


class _OptionTypeValidator:

    def __init__(self, data_types, value):
        self._data_types = data_types
        self._value = [value]

    @property
    def value(self):
        return self._value[-1]

    @contextmanager
    def push(self, value):
        self._value.append(value)
        try:
            yield
        finally:
            self._value.pop()

    def visit(self, type_):
        type_.accept(self)

    def visit_any(self, type_):
        pass

    def visit_boolean(self, type_):
        if not isinstance(self.value, bool):
            raise _OptionTypeError(self.value, type_)

    def visit_string(self, type_):
        if not isinstance(self.value, str):
            raise _OptionTypeError(self.value, type_)

    def visit_integer(self, type_):
        if not isinstance(self.value, int):
            raise _OptionTypeError(self.value, type_)

    def visit_float(self, type_):
        if not isinstance(self.value, float):
            raise _OptionTypeError(self.value, type_)

    def visit_optional(self, type_):
        if self.value is not None:
            self.visit(type_.__type__)

    def visit_sequence(self, type_):
        if not isinstance(self.value, collections_abc.Sequence):
            raise _OptionTypeError(self.value, type_)
        for item in self.value:
            with self.push(item):
                self.visit(type_.__item_type__)

    def visit_mapping(self, type_):
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)
        for key, value in self.value.items():
            with self.push(key):
                self.visit(type_.__key_type__)
            with self.push(value):
                self.visit(type_.__value_type__)

    def visit_record(self, type_):
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)

        unknown = set(self.value).difference(type_.__field_types__)
        if unknown:
            fields = ', '.join(sorted(map(repr, unknown)))
            raise _OptionError('unknown fields: {}'.format(fields))

        missing = set(type_.__field_types__).difference(self.value)
        if missing:
            fields = ', '.join(sorted(missing))
            raise _OptionError('missing fields: {}'.format(fields))

        for key, value_type in type_.__field_types__.items():
            with self.push(self.value[key]):
                self.visit(value_type)

    def visit_typeref(self, type_):
        assert type_.__type_name__ in self._data_types, type_.__type_name__
        self.visit(self._data_types[type_.__type_name__])


class _ValidateOptions(GraphVisitor):

    def __init__(self, data_types, options, for_, errors):
        self._data_types = data_types
        self.options = options
        self.for_ = for_
        self.errors = errors
        self._options = options or {}

    def visit_link(self, obj):
        super(_ValidateOptions, self).visit_link(obj)
        unknown = set(self._options).difference(obj.options_map)
        if unknown:
            node, field = self.for_
            self.errors.report('Unknown options for "{}.{}": {}'.
                               format(node, field, ', '.join(unknown)))

    visit_field = visit_link

    def visit_option(self, obj):
        value = self._options.get(obj.name, obj.default)
        if value is Nothing:
            node, field = self.for_
            self.errors.report('Required option "{}.{}:{}" is not specified'
                               .format(node, field, obj.name))
        elif obj.type is not None:
            try:
                _OptionTypeValidator(self._data_types, value).visit(obj.type)
            except _OptionError as err:
                node, field = self.for_
                self.errors.report('Invalid value for option "{}.{}:{}", {}'
                                   .format(node, field, obj.name,
                                           err.description))

    def visit_node(self, obj):
        assert self.options is None, 'Node can not have options'

    def visit_root(self, obj):
        raise AssertionError('Root node is not expected here')


class _RecordFieldsValidator(QueryVisitor):

    def __init__(self, data_types, field_types, errors):
        self._data_types = data_types
        self._field_types = field_types
        self._errors = errors

    def visit_field(self, obj):
        if obj.name not in self._field_types:
            self._errors.report('Unknown field name "{}"'.format(obj.name))
        elif obj.options is not None:
            self._errors.report('Options are not expected')
        elif _AssumeRecord(self._data_types).visit(self._field_types[obj.name]):
            self._errors.report('Trying to query "{}" link as it was a field'
                                .format(obj.name))

    def visit_link(self, obj):
        field_types = _AssumeRecord(self._data_types) \
            .visit(self._field_types[obj.name])
        if field_types is not None:
            fields_validator = _RecordFieldsValidator(self._data_types,
                                                      field_types,
                                                      self._errors)
            for field in obj.node.fields:
                fields_validator.visit(field)
        else:
            self._errors.report('"{}" is not a link'.format(obj.name))

    def visit_node(self, obj):
        raise AssertionError('Node is not expected here')


def _field_eq(a, b):
    return a.name == b.name and a.options == b.options


class QueryValidator(QueryVisitor):

    def __init__(self, graph):
        self.graph = graph
        self.path = [graph.root]
        self.errors = Errors()

    def visit_field(self, obj):
        node = self.path[-1]
        field = node.fields_map.get(obj.name)
        if field is not None:
            is_field = _AssumeField(node, self.errors).visit(field)
            if is_field:
                for_ = (node.name or 'root', obj.name)
                _ValidateOptions(self.graph.data_types, obj.options, for_,
                                 self.errors).visit(field)
        else:
            self.errors.report('Field "{}" is not implemented in the "{}" node'
                               .format(obj.name, node.name or 'root'))

    def visit_link(self, obj):
        node = self.path[-1]
        graph_obj = node.fields_map.get(obj.name, _undefined)
        if isinstance(graph_obj, Field):
            for_ = (node.name or 'root', obj.name)
            _ValidateOptions(self.graph.data_types, obj.options, for_,
                             self.errors).visit(graph_obj)

            field_types = _AssumeRecord(self.graph.data_types)\
                .visit(graph_obj.type)
            if field_types is not None:
                fields_validator = _RecordFieldsValidator(self.graph.data_types,
                                                          field_types,
                                                          self.errors)
                for field in obj.node.fields:
                    fields_validator.visit(field)
            else:
                self.errors.report('Trying to query "{}.{}" simple field '
                                   'as node'
                                   .format(node.name or 'root', obj.name))

        elif isinstance(graph_obj, Link):
            linked_node = self.graph.nodes_map[graph_obj.node]
            for_ = (node.name or 'root', obj.name)
            _ValidateOptions(self.graph.data_types, obj.options, for_,
                             self.errors).visit(graph_obj)

            self.path.append(linked_node)
            try:
                self.visit(obj.node)
            finally:
                self.path.pop()

        elif graph_obj is _undefined:
            self.errors.report('Link "{}" is not implemented in the "{}" node'
                               .format(obj.name, node.name or 'root'))
        else:
            raise TypeError(repr(graph_obj))

    def visit_node(self, obj):
        fields = {}
        for field in obj.fields:
            seen = fields.get(field.result_key)
            if seen is not None:
                if not _field_eq(field, seen):
                    node = self.path[-1].name or 'root'
                    self.errors.report('Found distinct fields with the same '
                                       'resulting name "{}" for the node "{}"'
                                       .format(field.result_key, node))
            else:
                fields[field.result_key] = field
            self.visit(field)


def validate(graph, query):
    query_validator = QueryValidator(graph)
    query_validator.visit(query)
    return query_validator.errors.list



from graphql.language import ast

from ..query import QueryVisitor, Field


def _name(value):
    return ast.NameNode(value=value) if value is not None else None


def _encode(value):
    if value is None:
        return ast.NullValueNode()
    elif isinstance(value, bool):
        return ast.BooleanValueNode(value=value)
    elif isinstance(value, int):
        return ast.IntValueNode(value=str(value))
    elif isinstance(value, float):
        return ast.FloatValueNode(value=str(value))
    elif isinstance(value, str):
        return ast.StringValueNode(value=value)
    elif isinstance(value, list):
        return ast.ListValueNode(values=[_encode(v) for v in value])
    elif isinstance(value, dict):
        return ast.ObjectValueNode(fields=[
            ast.ObjectFieldNode(name=_name(key), value=_encode(val))
            for key, val in value.items()
        ])
    else:
        raise TypeError('Unsupported type: {!r}'.format(value))


class Exporter(QueryVisitor):

    def visit_field(self, obj: Field):
        arguments = None
        if obj.options:
            arguments = [ast.ArgumentNode(name=_name(key), value=_encode(val))
                         for key, val in obj.options.items()]
        return ast.FieldNode(
            name=_name(obj.name),
            alias=_name(obj.alias),
            arguments=arguments,
        )

    def visit_link(self, obj):
        arguments = None
        if obj.options:
            arguments = [ast.ArgumentNode(name=_name(key), value=_encode(val))
                         for key, val in obj.options.items()]
        return ast.FieldNode(
            name=_name(obj.name),
            alias=_name(obj.alias),
            arguments=arguments,
            selection_set=self.visit(obj.node),
        )

    def visit_node(self, obj):
        return ast.SelectionSetNode(
            selections=[self.visit(f) for f in obj.fields],
        )


def export(query):
    return ast.DocumentNode(definitions=[
        ast.OperationDefinitionNode(
            operation=ast.OperationType.QUERY,
            selection_set=Exporter().visit(query),
        )
    ])



import re
import json
import typing as t

from dataclasses import dataclass
from functools import partial
from collections import OrderedDict

from ..directives import get_deprecated
from ..graph import Graph, Root, Node, Link, Option, Field, Nothing
from ..graph import GraphVisitor, GraphTransformer
from ..types import (
    TypeRef,
    String,
    Sequence,
    Boolean,
    Optional,
    TypeVisitor,
)
from ..types import Any, RecordMeta, AbstractTypeVisitor
from ..utils import (
    listify,
    cached_property,
)
from .types import (
    SCALAR,
    NON_NULL,
    LIST,
    INPUT_OBJECT,
    OBJECT,
    DIRECTIVE,
    FieldIdent,
    FieldArgIdent,
    InputObjectFieldIdent,
    DirectiveArgIdent,
)


@dataclass(frozen=True)
class Directive:
    @dataclass(frozen=True)
    class Argument:
        name: str
        type_ident: t.Any
        description: str
        default_value: t.Any

    name: str
    locations: t.List[str]
    description: str
    args: t.List[Argument]

    @property
    def args_map(self):
        return OrderedDict((arg.name, arg) for arg in self.args)


_BUILTIN_DIRECTIVES = (
    Directive(
        name='skip',
        locations=['FIELD', 'FRAGMENT_SPREAD', 'INLINE_FRAGMENT'],
        description=(
            'Directs the executor to skip this field or fragment '
            'when the `if` argument is true.'
        ),
        args=[
            Directive.Argument(
                name='if',
                type_ident=NON_NULL(SCALAR('Boolean')),
                description='Skipped when true.',
                default_value=None,
            ),
        ],
    ),
    Directive(
        name='include',
        locations=['FIELD', 'FRAGMENT_SPREAD', 'INLINE_FRAGMENT'],
        description=(
            'Directs the executor to include this field or fragment '
            'only when the `if` argument is true.'
        ),
        args=[
            Directive.Argument(
                name='if',
                type_ident=NON_NULL(SCALAR('Boolean')),
                description='Included when true.',
                default_value=None,
            ),
        ],
    ),
    Directive(
        name='deprecated',
        locations=['FIELD_DEFINITION', 'ENUM_VALUE'],
        description='Marks the field or enum value as deprecated',
        args=[
            Directive.Argument(
                name='reason',
                type_ident=SCALAR('String'),
                description='Deprecation reason.',
                default_value=None,
            ),
        ],
    ),
)


def _async_wrapper(func):
    async def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


QUERY_ROOT_NAME = 'Query'
MUTATION_ROOT_NAME = 'Mutation'


class SchemaInfo:

    def __init__(
        self,
        query_graph: Graph,
        mutation_graph: t.Optional[Graph] = None,
        directives: t.Optional[t.Sequence[Directive]] = None,
    ):
        self.query_graph = query_graph
        self.data_types = query_graph.data_types
        self.mutation_graph = mutation_graph
        self.directives = directives or ()

    @cached_property
    def directives_map(self):
        return OrderedDict((d.name, d) for d in self.directives)


class TypeIdent(AbstractTypeVisitor):

    def __init__(self, graph, input_mode=False):
        self._graph = graph
        self._input_mode = input_mode

    def visit_any(self, obj):
        return SCALAR('Any')

    def visit_mapping(self, obj):
        return SCALAR('Any')

    def visit_record(self, obj):
        return SCALAR('Any')

    def visit_callable(self, obj):
        raise TypeError('Not expected here: {!r}'.format(obj))

    def visit_sequence(self, obj):
        return NON_NULL(LIST(self.visit(obj.__item_type__)))

    def visit_optional(self, obj):
        ident = self.visit(obj.__type__)
        return ident.of_type if isinstance(ident, NON_NULL) else ident

    def visit_typeref(self, obj):
        if self._input_mode:
            assert obj.__type_name__ in self._graph.data_types, \
                obj.__type_name__
            return NON_NULL(INPUT_OBJECT(obj.__type_name__))
        else:
            return NON_NULL(OBJECT(obj.__type_name__))

    def visit_string(self, obj):
        return NON_NULL(SCALAR('String'))

    def visit_integer(self, obj):
        return NON_NULL(SCALAR('Int'))

    def visit_float(self, obj):
        return NON_NULL(SCALAR('Float'))

    def visit_boolean(self, obj):
        return NON_NULL(SCALAR('Boolean'))


class UnsupportedGraphQLType(TypeError):
    pass


class TypeValidator(TypeVisitor):

    @classmethod
    def is_valid(cls, type_):
        try:
            cls().visit(type_)
        except UnsupportedGraphQLType:
            return False
        else:
            return True

    def visit_any(self, obj):
        raise UnsupportedGraphQLType()

    def visit_record(self, obj):
        # inline Record type can't be directly matched to GraphQL type system
        raise UnsupportedGraphQLType()


def not_implemented(*args, **kwargs):
    raise NotImplementedError(args, kwargs)


def na_maybe(schema):
    return Nothing


def na_many(schema, ids=None, options=None):
    if ids is None:
        return []
    else:
        return [[] for _ in ids]


def _nodes_map(schema: SchemaInfo):
    nodes = [(n.name, n) for n in schema.query_graph.nodes]
    nodes.append((QUERY_ROOT_NAME, schema.query_graph.root))
    if schema.mutation_graph is not None:
        nodes.append((MUTATION_ROOT_NAME, schema.mutation_graph.root))
    return OrderedDict(nodes)


def schema_link(schema):
    return None


def type_link(schema, options):
    name = options['name']
    if name in _nodes_map(schema):
        return OBJECT(name)
    else:
        return Nothing


@listify
def root_schema_types(schema: SchemaInfo):
    yield SCALAR('String')
    yield SCALAR('Int')
    yield SCALAR('Boolean')
    yield SCALAR('Float')
    yield SCALAR('Any')

    for name in _nodes_map(schema):
        yield OBJECT(name)
    for name, type_ in schema.data_types.items():
        if isinstance(type_, RecordMeta):
            yield OBJECT(name)
            yield INPUT_OBJECT(name)


def root_schema_query_type(schema):
    return OBJECT(QUERY_ROOT_NAME)


def root_schema_mutation_type(schema):
    if schema.mutation_graph is not None:
        return OBJECT(MUTATION_ROOT_NAME)
    else:
        return Nothing


def root_schema_directives(schema):
    return [
        DIRECTIVE(directive.name) for directive in schema.directives
    ]


@listify
def type_info(schema, fields, ids):
    nodes_map = _nodes_map(schema)
    for ident in ids:
        if isinstance(ident, OBJECT):
            if ident.name in nodes_map:
                description = nodes_map[ident.name].description
            else:
                description = None
            info = {'id': ident,
                    'kind': 'OBJECT',
                    'name': ident.name,
                    'description': description}
        elif isinstance(ident, INPUT_OBJECT):
            info = {'id': ident,
                    'kind': 'INPUT_OBJECT',
                    'name': 'IO{}'.format(ident.name),
                    'description': None}
        elif isinstance(ident, NON_NULL):
            info = {'id': ident,
                    'kind': 'NON_NULL'}
        elif isinstance(ident, LIST):
            info = {'id': ident,
                    'kind': 'LIST'}
        elif isinstance(ident, SCALAR):
            info = {'id': ident,
                    'name': ident.name,
                    'kind': 'SCALAR'}
        else:
            raise TypeError(repr(ident))
        yield [info.get(f.name) for f in fields]


@listify
def type_fields_link(schema, ids, options):
    nodes_map = _nodes_map(schema)
    for ident in ids:
        if isinstance(ident, OBJECT):
            if ident.name in nodes_map:
                node = nodes_map[ident.name]
                field_idents = [
                    FieldIdent(ident.name, f.name)
                    for f in node.fields if not f.name.startswith('_')
                ]
            else:
                type_ = schema.data_types[ident.name]
                field_idents = [
                    FieldIdent(ident.name, f_name)
                    for f_name, f_type in type_.__field_types__.items()
                ]
            if not field_idents:
                raise TypeError('Object type "{}" does not contain fields, '
                                'which is not acceptable for GraphQL in order '
                                'to define schema type'.format(ident.name))
            yield field_idents
        else:
            yield []


@listify
def type_of_type_link(schema, ids):
    for ident in ids:
        if isinstance(ident, (NON_NULL, LIST)):
            yield ident.of_type
        else:
            yield Nothing


@listify
def field_info(schema, fields, ids):
    nodes_map = _nodes_map(schema)
    for ident in ids:
        if ident.node in nodes_map:
            node = nodes_map[ident.node]
            field = node.fields_map[ident.name]
            deprecated = None
            if isinstance(field, (Field, Link)):
                deprecated = get_deprecated(field)

            info = {'id': ident,
                    'name': field.name,
                    'description': field.description,
                    'isDeprecated': bool(deprecated),
                    'deprecationReason': deprecated and deprecated.reason}
        else:
            info = {'id': ident,
                    'name': ident.name,
                    'description': None,
                    'isDeprecated': False,
                    'deprecationReason': None}
        yield [info[f.name] for f in fields]


@listify
def field_type_link(schema, ids):
    nodes_map = _nodes_map(schema)
    type_ident = TypeIdent(schema.query_graph)
    for ident in ids:
        if ident.node in nodes_map:
            node = nodes_map[ident.node]
            field = node.fields_map[ident.name]
            yield type_ident.visit(field.type or Any)
        else:
            data_type = schema.data_types[ident.node]
            field_type = data_type.__field_types__[ident.name]
            yield type_ident.visit(field_type)


@listify
def field_args_link(schema, ids):
    nodes_map = _nodes_map(schema)
    for ident in ids:
        if ident.node in nodes_map:
            node = nodes_map[ident.node]
            field = node.fields_map[ident.name]
            yield [FieldArgIdent(ident.node, field.name, option.name)
                   for option in field.options]
        else:
            yield []


@listify
def type_input_object_input_fields_link(schema, ids):
    for ident in ids:
        if isinstance(ident, INPUT_OBJECT):
            data_type = schema.data_types[ident.name]
            yield [InputObjectFieldIdent(ident.name, key)
                   for key in data_type.__field_types__.keys()]
        else:
            yield []


@listify
def input_value_info(schema, fields, ids):
    nodes_map = _nodes_map(schema)
    for ident in ids:
        if isinstance(ident, FieldArgIdent):
            node = nodes_map[ident.node]
            field = node.fields_map[ident.field]
            option = field.options_map[ident.name]
            if option.default is Nothing:
                default = None
            else:
                default = json.dumps(option.default)
            info = {'id': ident,
                    'name': option.name,
                    'description': option.description,
                    'defaultValue': default}
            yield [info[f.name] for f in fields]
        elif isinstance(ident, InputObjectFieldIdent):
            info = {'id': ident,
                    'name': ident.key,
                    'description': None,
                    'defaultValue': None}
            yield [info[f.name] for f in fields]
        elif isinstance(ident, DirectiveArgIdent):
            directive = schema.directives_map[ident.name]
            arg = directive.args_map[ident.arg]
            info = {'id': ident,
                    'name': arg.name,
                    'description': arg.description,
                    'defaultValue': arg.default_value}
            yield [info[f.name] for f in fields]
        else:
            raise TypeError(repr(ident))


@listify
def input_value_type_link(schema, ids):
    nodes_map = _nodes_map(schema)
    type_ident = TypeIdent(schema.query_graph, input_mode=True)
    for ident in ids:
        if isinstance(ident, FieldArgIdent):
            node = nodes_map[ident.node]
            field = node.fields_map[ident.field]
            option = field.options_map[ident.name]
            yield type_ident.visit(option.type)
        elif isinstance(ident, InputObjectFieldIdent):
            data_type = schema.data_types[ident.name]
            field_type = data_type.__field_types__[ident.key]
            yield type_ident.visit(field_type)
        elif isinstance(ident, DirectiveArgIdent):
            directive = schema.directives_map[ident.name]
            for arg in directive.args:
                yield arg.type_ident
        else:
            raise TypeError(repr(ident))


@listify
def directive_value_info(schema, fields, ids):
    for ident in ids:
        if ident.name in schema.directives_map:
            directive = schema.directives_map[ident.name]
            info = {'name': directive.name,
                    'description': directive.description,
                    'locations': directive.locations}
            yield [info[f.name] for f in fields]


def directive_args_link(schema, ids):
    links = []
    for ident in ids:
        directive = schema.directives_map[ident]
        links.append([DirectiveArgIdent(ident, arg.name)
                      for arg in directive.args])
    return links


GRAPH = Graph([
    Node('__Type', [
        Field('id', None, type_info),
        Field('kind', String, type_info),
        Field('name', String, type_info),
        Field('description', String, type_info),

        # OBJECT and INTERFACE only
        Link('fields', Sequence[TypeRef['__Field']], type_fields_link,
             requires='id',
             options=[Option('includeDeprecated', Boolean, default=False)]),

        # OBJECT only
        Link('interfaces', Sequence[TypeRef['__Type']], na_many,
             requires='id'),

        # INTERFACE and UNION only
        Link('possibleTypes', Sequence[TypeRef['__Type']], na_many,
             requires='id'),

        # ENUM only
        Link('enumValues', Sequence[TypeRef['__EnumValue']], na_many,
             requires='id',
             options=[Option('includeDeprecated', Boolean, default=False)]),

        # INPUT_OBJECT only
        Link('inputFields', Sequence[TypeRef['__InputValue']],
             type_input_object_input_fields_link, requires='id'),

        # NON_NULL and LIST only
        Link('ofType', Optional[TypeRef['__Type']], type_of_type_link,
             requires='id'),
    ]),
    Node('__Field', [
        Field('id', None, field_info),
        Field('name', String, field_info),
        Field('description', String, field_info),

        Link('args', Sequence[TypeRef['__InputValue']], field_args_link,
             requires='id'),
        Link('type', TypeRef['__Type'], field_type_link, requires='id'),
        Field('isDeprecated', Boolean, field_info),
        Field('deprecationReason', String, field_info),
    ]),
    Node('__InputValue', [
        Field('id', None, input_value_info),
        Field('name', String, input_value_info),
        Field('description', String, input_value_info),
        Link('type', TypeRef['__Type'], input_value_type_link, requires='id'),
        Field('defaultValue', String, input_value_info),
    ]),
    Node('__Directive', [
        Field('name', String, directive_value_info),
        Field('description', String, directive_value_info),
        Field('locations', Sequence[String], directive_value_info),
        Link('args', Sequence[TypeRef['__InputValue']], directive_args_link,
             requires='name'),
    ]),
    Node('__EnumValue', [
        Field('name', String, not_implemented),
        Field('description', String, not_implemented),
        Field('isDeprecated', Boolean, not_implemented),
        Field('deprecationReason', String, not_implemented),
    ]),
    Node('__Schema', [
        Link('types', Sequence[TypeRef['__Type']], root_schema_types,
             requires=None),
        Link('queryType', TypeRef['__Type'],
             root_schema_query_type, requires=None),
        Link('mutationType', Optional[TypeRef['__Type']],
             root_schema_mutation_type, requires=None),
        Link('subscriptionType', Optional[TypeRef['__Type']], na_maybe,
             requires=None),
        Link('directives', Sequence[TypeRef['__Directive']],
             root_schema_directives, requires=None),
    ]),
    Root([
        Link('__schema', TypeRef['__Schema'], schema_link, requires=None),
        Link('__type', Optional[TypeRef['__Type']], type_link, requires=None,
             options=[Option('name', String)]),
    ]),
])


class ValidateGraph(GraphVisitor):
    _name_re = re.compile(r'^[_a-zA-Z]\w*$', re.ASCII)

    def __init__(self):
        self._path = []
        self._errors = []

    def _add_error(self, name, description):
        path = '.'.join(self._path + [name])
        self._errors.append('{}: {}'.format(path, description))

    @classmethod
    def validate(cls, graph):
        self = cls()
        self.visit(graph)
        if self._errors:
            raise ValueError('Invalid GraphQL graph:\n{}'
                             .format('\n'.join('- {}'.format(err)
                                               for err in self._errors)))

    def visit_node(self, obj):
        if not self._name_re.match(obj.name):
            self._add_error(obj.name,
                            'Invalid node name: {}'.format(obj.name))
        if obj.fields:
            self._path.append(obj.name)
            super(ValidateGraph, self).visit_node(obj)
            self._path.pop()
        else:
            self._add_error(obj.name,
                            'No fields in the {} node'.format(obj.name))

    def visit_root(self, obj):
        if obj.fields:
            self._path.append('Root')
            super(ValidateGraph, self).visit_root(obj)
            self._path.pop()
        else:
            self._add_error('Root', 'No fields in the Root node')

    def visit_field(self, obj):
        if not self._name_re.match(obj.name):
            self._add_error(obj.name,
                            'Invalid field name: {}'.format(obj.name))
        super(ValidateGraph, self).visit_field(obj)

    def visit_link(self, obj):
        if not self._name_re.match(obj.name):
            self._add_error(obj.name,
                            'Invalid link name: {}'.format(obj.name))
        super(ValidateGraph, self).visit_link(obj)

    def visit_option(self, obj):
        if not self._name_re.match(obj.name):
            self._add_error(obj.name,
                            'Invalid option name: {}'.format(obj.name))
        super(ValidateGraph, self).visit_option(obj)


class BindToSchema(GraphTransformer):

    def __init__(self, schema):
        self.schema = schema
        self._processed = {}

    def visit_field(self, obj):
        field = super(BindToSchema, self).visit_field(obj)
        func = self._processed.get(obj.func)
        if func is None:
            func = self._processed[obj.func] = partial(obj.func, self.schema)
        field.func = func
        return field

    def visit_link(self, obj):
        link = super(BindToSchema, self).visit_link(obj)
        link.func = partial(link.func, self.schema)
        return link


class MakeAsync(GraphTransformer):

    def __init__(self):
        self._processed = {}

    def visit_field(self, obj):
        field = super(MakeAsync, self).visit_field(obj)
        func = self._processed.get(obj.func)
        if func is None:
            func = self._processed[obj.func] = _async_wrapper(obj.func)
        field.func = func
        return field

    def visit_link(self, obj):
        link = super(MakeAsync, self).visit_link(obj)
        link.func = _async_wrapper(link.func)
        return link


def type_name_field_func(node_name, fields, ids=None):
    return [[node_name] for _ in ids] if ids is not None else [node_name]


class AddIntrospection(GraphTransformer):

    def __init__(self, introspection_graph, type_name_field_factory):
        self.introspection_graph = introspection_graph
        self.type_name_field_factory = type_name_field_factory

    def visit_node(self, obj):
        node = super(AddIntrospection, self).visit_node(obj)
        node.fields.append(self.type_name_field_factory(obj.name))
        return node

    def visit_root(self, obj):
        root = super(AddIntrospection, self).visit_root(obj)
        root.fields.append(self.type_name_field_factory(QUERY_ROOT_NAME))
        return root

    def visit_graph(self, obj):
        graph = super(AddIntrospection, self).visit_graph(obj)
        graph.items.extend(self.introspection_graph.items)
        return graph


class GraphQLIntrospection(GraphTransformer):
    """Adds GraphQL introspection into synchronous graph

    Example:

    .. code-block:: python

        from hiku.graph import apply
        from hiku.introspection.graphql import GraphQLIntrospection

        graph = apply(graph, [GraphQLIntrospection(graph)])

    """
    __directives__: t.Tuple[Directive, ...] = _BUILTIN_DIRECTIVES

    def __init__(self, query_graph, mutation_graph=None):
        """
        :param query_graph: graph, where Root node represents Query root
            operation type
        :param mutation_graph: graph, where Root node represents Mutation root
            operation type
        """
        self._schema = SchemaInfo(
            query_graph,
            mutation_graph,
            self.__directives__,
        )

    def __type_name__(self, node_name):
        return Field('__typename', String,
                     partial(type_name_field_func, node_name))

    def __introspection_graph__(self):
        return BindToSchema(self._schema).visit(GRAPH)

    def visit_node(self, obj):
        node = super(GraphQLIntrospection, self).visit_node(obj)
        node.fields.append(self.__type_name__(obj.name))
        return node

    def visit_root(self, obj):
        root = super(GraphQLIntrospection, self).visit_root(obj)
        root.fields.append(self.__type_name__(QUERY_ROOT_NAME))
        return root

    def visit_graph(self, obj):
        ValidateGraph.validate(obj)
        introspection_graph = self.__introspection_graph__()
        items = [self.visit(node) for node in obj.items]
        items.extend(introspection_graph.items)
        return Graph(items, data_types=obj.data_types)


class AsyncGraphQLIntrospection(GraphQLIntrospection):
    """Adds GraphQL introspection into asynchronous graph

    Example:

    .. code-block:: python

        from hiku.graph import apply
        from hiku.introspection.graphql import AsyncGraphQLIntrospection

        graph = apply(graph, [AsyncGraphQLIntrospection(graph)])

    """
    def __type_name__(self, node_name):
        return Field('__typename', String,
                     _async_wrapper(partial(type_name_field_func, node_name)))

    def __introspection_graph__(self):
        graph = super(AsyncGraphQLIntrospection, self).__introspection_graph__()
        graph = MakeAsync().visit(graph)
        return graph



from ..graph import Graph

from ..introspection.types import NON_NULL, SCALAR
from ..introspection.graphql import Directive, GraphQLIntrospection
from ..introspection.graphql import AsyncGraphQLIntrospection

from .utils import get_keys


_DIRECTIVES = (
    # https://www.apollographql.com/docs/federation/federation-spec/#key
    Directive(
        name='key',
        locations=['OBJECT', 'INTERFACE'],
        description=(
            'The @key directive is used to indicate a combination '
            'of fields that can be used to uniquely identify and '
            'fetch an object or interface.'
        ),
        args=[
            Directive.Argument(
                name='fields',
                type_ident=NON_NULL(SCALAR('_FieldSet')),
                description=(
                    'A combination of fields that can be used to uniquely '
                    'identify and fetch an object or interface'
                ),
                default_value=None,
            ),
        ],
    ),
    # https://www.apollographql.com/docs/federation/federation-spec/#provides
    Directive(
        name='provides',
        locations=['FIELD_DEFINITION'],
        description=(
            'The @provides directive is used to annotate the expected returned '
            'fieldset from a field on a base type that is guaranteed to be '
            'selectable by the gateway'
        ),
        args=[
            Directive.Argument(
                name='fields',
                type_ident=NON_NULL(SCALAR('_FieldSet')),
                description=(
                    'Expected returned fieldset from a field on a base type '
                    'that is guaranteed to be selectable by the gateway'
                ),
                default_value=None,
            ),
        ],
    ),
    # https://www.apollographql.com/docs/federation/federation-spec/#requires
    Directive(
        name='requires',
        locations=['FIELD_DEFINITION'],
        description=(
            'The @requires directive is used to annotate the required input '
            'fieldset from a base type for a resolver.'
        ),
        args=[
            Directive.Argument(
                name='fields',
                type_ident=NON_NULL(SCALAR('_FieldSet')),
                description=(
                    'The required input fieldset from a base type for a '
                    'resolver'
                ),
                default_value=None,
            ),
        ],
    ),
    # https://www.apollographql.com/docs/federation/federation-spec/#external
    Directive(
        name='external',
        locations=['FIELD_DEFINITION'],
        description=(
            'The @external directive is used to mark a field '
            'as owned by another service.'
        ),
        args=[],
    ),
)


def is_introspection_query(query):
    return '__schema' in query.fields_map


def _obj(name):
    return {'kind': 'OBJECT', 'name': name, 'ofType': None}


def _non_null(t):
    return {'kind': 'NON_NULL', 'name': None, 'ofType': t}


def _union(name, possible_types=None):
    return {
        'kind': 'UNION',
        'name': name,
        'possibleTypes': possible_types
    }


def _field(name, type_, **kwargs):
    data = {
        'args': [],
        'deprecationReason': None,
        'description': None,
        'isDeprecated': False,
        'name': name,
        'type': type_
    }
    data.update(kwargs)
    return data


def _seq_of_nullable(_type):
    return {'kind': 'NON_NULL', 'name': None,
            'ofType': {'kind': 'LIST', 'name': None,
                       'ofType': _type}}


def _seq_of(_type):
    return _seq_of_nullable({'kind': 'NON_NULL', 'name': None,
                            'ofType': _type})


def _ival(name, type_, **kwargs):
    data = {
        'name': name,
        'type': type_,
        'description': None,
        'defaultValue': None,
    }
    data.update(kwargs)
    return data


def _type(name, kind, **kwargs):
    data = {
        'description': None,
        'enumValues': [],
        'fields': [],
        'inputFields': [],
        'interfaces': [],
        'kind': kind,
        'name': name,
        'possibleTypes': [],
    }
    data.update(**kwargs)
    return data


def extend_with_federation(graph: Graph, data: dict):
    union_types = []
    for node in graph.nodes:
        if get_keys(graph, node.name):
            union_types.append(_obj(node.name))

    data['__schema']['types'].append(_type('_Any', 'SCALAR'))
    data['__schema']['types'].append(_type('_FieldSet', 'SCALAR'))
    data['__schema']['types'].append(
        _type('_Entity', 'UNION', possibleTypes=union_types)
    )
    data['__schema']['types'].append(
        _type('_Service', 'OBJECT',
              fields=[_field('sdl', _type('String', 'SCALAR'))])
    )

    entities_field = _field(
        '_entities',
        _seq_of_nullable(_union('_Entity', union_types)),
        args=[_ival('representations', _seq_of(_type('_Any', 'SCALAR')))]
    )

    service_field = _field(
        '_service',
        _non_null(_obj('_Service')),
    )
    for t in data['__schema']['types']:
        if t['kind'] == 'OBJECT' and t['name'] == 'Query':
            t['fields'].append(entities_field)
            t['fields'].append(service_field)


class FederatedGraphQLIntrospection(GraphQLIntrospection):
    """
    Federation-aware introspection
    https://www.apollographql.com/docs/federation/federation-spec/#federation-schema-specification
    """
    __directives__ = GraphQLIntrospection.__directives__ + _DIRECTIVES


class AsyncFederatedGraphQLIntrospection(
    FederatedGraphQLIntrospection,
    AsyncGraphQLIntrospection
):
    """Adds GraphQL introspection into asynchronous federated graph"""



import re
import os.path

from setuptools import setup, find_packages


with open(
    os.path.join(os.path.dirname(__file__), 'hiku', '__init__.py')
) as f:
    VERSION = re.match(r".*__version__ = '(.*?)'", f.read(), re.S).group(1)

with open(
    os.path.join(os.path.dirname(__file__), 'README.rst')
) as f:
    DESCRIPTION = f.read()

setup(
    name='hiku',
    version=VERSION,
    description='Library to implement Graph APIs',
    long_description=DESCRIPTION,
    long_description_content_type='text/x-rst',
    author='Vladimir Magamedov',
    author_email='vladimir@magamedov.com',
    url='https://github.com/evo-company/hiku',
    packages=find_packages(exclude=['test*']),
    include_package_data=True,
    license='BSD-3-Clause',
    python_requires='>=3.7',
    install_requires=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



"""
    hiku.graph
    ~~~~~~~~~~

    Graphs are defined by nodes, fields and links. Simple functions
    are used to fetch any data from any data source.

"""
import typing as t

from abc import ABC, abstractmethod
from itertools import chain
from functools import reduce
from collections import OrderedDict
from typing import List

from .types import (
    OptionalMeta,
    SequenceMeta,
    TypeRefMeta,
    Record,
    Any,
    GenericMeta,
)
from .utils import cached_property, const


Maybe = const('Maybe')

One = const('One')

Many = const('Many')

#: Special constant that is used by links with :py:class:`~hiku.types.Optional`
#: type in order to indicate that there is nothing to link to
Nothing = const('Nothing')


class AbstractBase(ABC):

    @abstractmethod
    def accept(self, visitor):
        pass


class AbstractOption(AbstractBase):
    pass


class Option(AbstractOption):
    """Defines an option, used to customize results of the fields and links

    Options without default value are **required**.

    Example of a required option::

        Option('id', Integer)

    Example of an optional option::

        Option('size', Integer, default=100)

    """
    def __init__(
        self,
        name: str,
        type_: GenericMeta,
        *,
        default: t.Any = Nothing,
        description: t.Optional[str] = None
    ):
        """
        :param name: name of the option
        :param type_: type of the option or ``None``
        :param default: default option value
        :param description: description of the option
        """
        self.name = name
        self.type = type_
        self.default = default
        self.description = description

    def __repr__(self):
        return '{}({!r}, {!r}, ...)'.format(self.__class__.__name__,
                                            self.name, self.type)

    def accept(self, visitor):
        return visitor.visit_option(self)


class AbstractField(AbstractBase):
    pass


class Field(AbstractField):
    """Defines a field of the node

    Example::

        graph = Graph([
            Node('user', [
                Field('name', String, func),
            ]),
        ])

    Example with options::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)]),
            ]),
        ])

    Example with directives::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)],
                      directives=[Deprecated('use another field')]),
            ]),
        ])

    Data loading protocol::

        # root node fields
        def func(fields) -> List[T]

        # non-root node fields
        def func(fields, ids) -> List[List[T]]

    Where:

    - ``fields`` - list of :py:class:`hiku.query.Field`
    - ``ids`` - list node identifiers

    """
    def __init__(self, name, type_, func, *, options=None, description=None,
                 directives=None):
        """
        :param str name: name of the field
        :param type_: type of the field or ``None``
        :param func: function to load field's data
        :param options: list of acceptable options
        :param description: description of the field
        :param directives: list of directives for the field
        """
        self.name = name
        self.type = type_
        self.func = func
        self.options = options or ()
        self.description = description
        self.directives = directives or ()

    def __repr__(self):
        return '{}({!r}, {!r}, {!r})'.format(self.__class__.__name__, self.name,
                                             self.type, self.func)

    @cached_property
    def options_map(self):
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor):
        return visitor.visit_field(self)


class AbstractLink(AbstractBase):
    pass


def get_type_enum(type_):
    if isinstance(type_, TypeRefMeta):
        return One, type_.__type_name__
    elif isinstance(type_, OptionalMeta):
        if isinstance(type_.__type__, TypeRefMeta):
            return Maybe, type_.__type__.__type_name__
    elif isinstance(type_, SequenceMeta):
        if isinstance(type_.__item_type__, TypeRefMeta):
            return Many, type_.__item_type__.__type_name__
    raise TypeError('Invalid type specified: {!r}'.format(type_))


class Link(AbstractLink):
    """Defines a link to the node

    Example::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None),
            ]),
        ])

    Example with requirements::

        graph = Graph([
            Node('character', [...]),
            Node('actor', [
                Field('id', Integer, field_func),
                Link('characters', Sequence[TypeRef['character']],
                     link_func, requires='id'),
            ])
        ])

    Requirements are needed when link points from non-root node.

    Example with options::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)]),
            ]),
        ])

    Example with directives::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)],
                     directives=[Deprecated('do not use')]),
            ]),
        ])

    Identifiers loading function protocol::

        # From root node or if requires is None:

        # ... if type is TypeRef['foo']
        def func() -> T

        # ... if type is Optional[TypeRef['foo']]
        def func() -> Union[T, Nothing]

        # ... if type is Sequence[TypeRef['foo']]
        def func() -> List[T]

        # From non-root node and requires is not None:

        # ... if type is TypeRef['foo']
        def func(ids) -> List[T]

        # ... if type is Optional[TypeRef['foo']]
        def func(ids) -> List[Union[T, Nothing]]

        # ... if type is Sequence[TypeRef['foo']]
        def func(ids) -> List[List[T]]

    See also :py:const:`hiku.graph.Nothing`.

    If link was defined with options, then link function should accept one
    additional positional argument::

        # many to many link with options
        def func(ids, options) -> List[List[T]]

    Where ``options`` is a mapping ``str: value`` of provided in the query
    options.
    """
    def __init__(
        self, name, type_, func, *, requires, options=None, description=None,
        directives=None
    ):
        """
        :param name: name of the link
        :param type_: type of the link
        :param func: function to load identifiers of the linked node
        :param requires: field name from the current node, required to compute
                         identifiers of the linked node
        :param options: list of acceptable options
        :param description: description of the link
        :param directives: list of directives for the link
        """
        type_enum, node = get_type_enum(type_)

        self.name = name
        self.type = type_
        self.type_enum = type_enum
        self.node = node
        self.func = func
        self.requires = requires
        self.options = options or ()
        self.description = description
        self.directives = directives or ()

    def __repr__(self):
        return '{}({!r}, {!r}, {!r}, ...)'.format(self.__class__.__name__,
                                                  self.name, self.type,
                                                  self.func)

    @cached_property
    def options_map(self):
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor):
        return visitor.visit_link(self)


class AbstractNode(AbstractBase):
    pass


class Node(AbstractNode):
    """Collection of the fields and links, which describes some entity and
    relations with other entities

    Example::

        graph = Graph([
            Node('user', [
                Field('id', Integer, field_func),
                Field('name', String, field_func),
                Link('roles', Sequence[TypeRef['role']],
                     link_func, requires='id'),
            ]),
        ])

    """
    def __init__(self, name, fields, *, description=None, directives=None):
        """
        :param name: name of the node
        :param fields: list of fields and links
        :param description: description of the node
        """
        self.name = name
        self.fields = fields
        self.description = description
        self.directives = directives or ()

    def __repr__(self):
        return '{}({!r}, {!r}, ...)'.format(self.__class__.__name__, self.name,
                                            self.fields)

    @cached_property
    def fields_map(self):
        return OrderedDict((f.name, f) for f in self.fields)

    def accept(self, visitor):
        return visitor.visit_node(self)


class Root(Node):
    """Special implicit root node, starting point of the query execution

    Example::

        graph = Graph([
            Node('baz', [...]),
            Root([
                Field('foo', String, root_fields_func),
                Link('bar', Sequence[TypeRef['baz']],
                     to_baz_func, requires=None),
            ]),
        ])

    """
    def __init__(self, items):
        """
        :param items: list of fields, links and singleton nodes
        """
        super(Root, self).__init__(None, items)

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__, self.name,
                                       self.fields)

    def accept(self, visitor):
        return visitor.visit_root(self)


class AbstractGraph(AbstractBase):
    pass


class Graph(AbstractGraph):
    """Collection of nodes - definition of the graph

    Example::

        graph = Graph([
            Node('foo', [...]),
            Node('bar', [...]),
            Root([...]),
        ])

    """
    def __init__(self, items, data_types=None):
        """
        :param items: list of nodes
        """
        from .validate.graph import GraphValidator

        GraphValidator.validate(items)

        self.items = GraphInit.init(items)
        self.data_types = data_types or {}
        self.__types__ = GraphTypes.get_types(self.items, self.data_types)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.items)

    def iter_root(self):
        for node in self.items:
            if node.name is None:
                for field in node.fields:
                    yield field

    def iter_nodes(self):
        for node in self.items:
            if node.name is not None:
                yield node

    @cached_property
    def root(self):
        return Root(list(self.iter_root()))

    @cached_property
    def nodes(self):
        return list(self.iter_nodes())

    @cached_property
    def nodes_map(self):
        return OrderedDict((e.name, e) for e in self.iter_nodes())

    def accept(self, visitor):
        return visitor.visit_graph(self)


class AbstractGraphVisitor(ABC):

    @abstractmethod
    def visit(self, obj):
        pass

    @abstractmethod
    def visit_option(self, obj):
        pass

    @abstractmethod
    def visit_field(self, obj):
        pass

    @abstractmethod
    def visit_link(self, obj):
        pass

    @abstractmethod
    def visit_node(self, obj):
        pass

    @abstractmethod
    def visit_root(self, obj):
        pass

    @abstractmethod
    def visit_graph(self, obj):
        pass


class GraphVisitor(AbstractGraphVisitor):

    def visit(self, obj):
        return obj.accept(self)

    def visit_option(self, obj):
        pass

    def visit_field(self, obj):
        for option in obj.options:
            self.visit(option)

    def visit_link(self, obj):
        for option in obj.options:
            self.visit(option)

    def visit_node(self, obj):
        for item in obj.fields:
            self.visit(item)

    def visit_root(self, obj):
        for item in obj.fields:
            self.visit(item)

    def visit_graph(self, obj):
        for item in obj.items:
            self.visit(item)


class GraphTransformer(AbstractGraphVisitor):

    def visit(self, obj):
        return obj.accept(self)

    def visit_option(self, obj):
        return Option(obj.name, obj.type, default=obj.default,
                      description=obj.description)

    def visit_field(self, obj):
        return Field(obj.name, obj.type, obj.func,
                     options=[self.visit(op) for op in obj.options],
                     description=obj.description, directives=obj.directives)

    def visit_link(self, obj):
        return Link(obj.name, obj.type, obj.func,
                    requires=obj.requires,
                    options=[self.visit(op) for op in obj.options],
                    description=obj.description, directives=obj.directives)

    def visit_node(self, obj):
        return Node(obj.name, [self.visit(f) for f in obj.fields],
                    description=obj.description, directives=obj.directives)

    def visit_root(self, obj):
        return Root([self.visit(f) for f in obj.fields])

    def visit_graph(self, obj):
        return Graph([self.visit(node) for node in obj.items],
                     obj.data_types)


def apply(graph: Graph, transformers: List[GraphTransformer]) -> Graph:
    """Helper function to apply graph transformations

    Example:

    .. code-block:: python

        graph = hiku.graph.apply(graph, [AsyncGraphQLIntrospection()])

    """
    return reduce(lambda g, t: t.visit(g), transformers, graph)


class GraphInit(GraphTransformer):

    @classmethod
    def init(cls, items):
        self = cls()
        return [self.visit(i) for i in items]

    def visit_field(self, obj):
        field = super(GraphInit, self).visit_field(obj)
        postprocess = getattr(field.func, '__postprocess__', None)
        if postprocess is not None:
            postprocess(field)
        return field

    def visit_link(self, obj):
        link = super(GraphInit, self).visit_link(obj)
        postprocess = getattr(link.func, '__postprocess__', None)
        if postprocess is not None:
            postprocess(link)
        return link


class GraphTypes(GraphVisitor):

    def _visit_graph(self, items, data_types):
        types = OrderedDict(data_types)
        roots = []
        for item in items:
            if item.name is not None:
                types[item.name] = self.visit(item)
            else:
                roots.append(self.visit(item))
        types['__root__'] = Record[chain.from_iterable(
            r.__field_types__.items() for r in roots
        )]
        return types

    @classmethod
    def get_types(cls, items, data_types):
        return cls()._visit_graph(items, data_types)

    def visit_graph(self, obj):
        return self._visit_graph(obj.items, obj.data_types)

    def visit_node(self, obj):
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_root(self, obj):
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_link(self, obj):
        return obj.type

    def visit_field(self, obj):
        return obj.type or Any



from functools import partial
from collections import defaultdict

import sqlalchemy

from ..types import String, Integer
from ..graph import Nothing, Maybe, One, Many
from ..engine import pass_context


def _translate_type(column):
    if isinstance(column.type, sqlalchemy.Integer):
        return Integer
    elif isinstance(column.type, sqlalchemy.Unicode):
        return String
    else:
        return None


def _table_repr(table):
    return 'Table({})'.format(', '.join(
        [repr(table.name), repr(table.metadata), '...',
         'schema={!r}'.format(table.schema)]
    ))


@pass_context
class FieldsQuery:

    def __init__(self, engine_key, from_clause, *, primary_key=None):
        self.engine_key = engine_key
        self.from_clause = from_clause
        if primary_key is not None:
            self.primary_key = primary_key
        else:
            # currently only one column supported
            self.primary_key, = from_clause.primary_key

    def __repr__(self):
        if isinstance(self.from_clause, sqlalchemy.Table):
            from_clause_repr = _table_repr(self.from_clause)
        else:
            from_clause_repr = repr(self.from_clause)
        return ('<{}.{}: engine_key={!r}, from_clause={}, primary_key={!r}>'
                .format(self.__class__.__module__, self.__class__.__name__,
                        self.engine_key, from_clause_repr, self.primary_key))

    def __postprocess__(self, field):
        if field.type is None:
            column = self.from_clause.c[field.name]
            field.type = _translate_type(column)

    def in_impl(self, column, values):
        return column.in_(values)

    def select_expr(self, fields_, ids):
        columns = [self.from_clause.c[f.name] for f in fields_]
        expr = (
            sqlalchemy.select([self.primary_key] + columns)
            .select_from(self.from_clause)
            .where(self.in_impl(self.primary_key, ids))
        )

        def result_proc(rows):
            rows_map = {row[self.primary_key]: [row[c] for c in columns]
                        for row in rows}

            nulls = [None for _ in fields_]
            return [rows_map.get(id_, nulls) for id_ in ids]

        return expr, result_proc

    def __call__(self, ctx, fields_, ids):
        if not ids:
            return []

        expr, result_proc = self.select_expr(fields_, ids)

        sa_engine = ctx[self.engine_key]
        with sa_engine.connect() as connection:
            rows = connection.execute(expr).fetchall()

        return result_proc(rows)


def _to_maybe_mapper(pairs, values):
    mapping = dict(pairs)
    return [mapping.get(value, Nothing) for value in values]


def _to_one_mapper(pairs, values):
    mapping = dict(pairs)
    return [mapping[value] for value in values]


def _to_many_mapper(pairs, values):
    mapping = defaultdict(list)
    for from_value, to_value in pairs:
        mapping[from_value].append(to_value)
    return [mapping[value] for value in values]


class LinkQuery:

    def __init__(self, engine_key, *, from_column, to_column):
        if from_column.table is not to_column.table:
            raise ValueError('from_column and to_column should belong to '
                             'one table')

        self.engine_key = engine_key
        self.from_column = from_column
        self.to_column = to_column

    def __repr__(self):
        return ('<{}.{}: engine_key={!r}, from_column={!r}, to_column={!r}>'
                .format(self.__class__.__module__, self.__class__.__name__,
                        self.engine_key, self.from_column, self.to_column))

    def __postprocess__(self, link):
        if link.type_enum is One:
            func = partial(self, _to_one_mapper)
        elif link.type_enum is Maybe:
            func = partial(self, _to_maybe_mapper)
        elif link.type_enum is Many:
            func = partial(self, _to_many_mapper)
        else:
            raise TypeError(repr(link.type_enum))
        link.func = pass_context(func)

    def in_impl(self, column, values):
        return column.in_(values)

    def select_expr(self, ids):
        # TODO: make this optional, but enabled by default
        filtered_ids = [i for i in set(ids) if i is not None]
        if filtered_ids:
            return (
                sqlalchemy.select([self.from_column.label('from_column'),
                                   self.to_column.label('to_column')])
                .where(self.in_impl(self.from_column, filtered_ids))
            )
        else:
            return None

    def __call__(self, result_proc, ctx, ids):
        expr = self.select_expr(ids)
        if expr is None:
            pairs = []
        else:
            sa_engine = ctx[self.engine_key]
            with sa_engine.connect() as connection:
                pairs = connection.execute(expr).fetchall()
        return result_proc(pairs, ids)
